# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v0_messages.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11v0_messages.proto\x1a\x1egoogle/protobuf/wrappers.proto\"\x81\x01\n\nCountersV0\x12\r\n\x05\x63ount\x18\x01 \x01(\x03\x12/\n\ntrue_count\x18\x02 \x01(\x0b\x32\x1b.google.protobuf.Int64Value\x12\x33\n\nnull_count\x18\x03 \x01(\x0b\x32\x1b.google.protobuf.Int64ValueB\x02\x18\x01\"\x95\x01\n\x0cInferredType\x12 \n\x04type\x18\x01 \x01(\x0e\x32\x12.InferredType.Type\x12\r\n\x05ratio\x18\x02 \x01(\x01\"T\n\x04Type\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x08\n\x04NULL\x10\x01\x12\x0e\n\nFRACTIONAL\x10\x02\x12\x0c\n\x08INTEGRAL\x10\x03\x12\x0b\n\x07\x42OOLEAN\x10\x04\x12\n\n\x06STRING\x10\x05\"F\n\x0e\x44oublesMessage\x12\r\n\x05\x63ount\x18\x01 \x01(\x03\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\x12\x0b\n\x03sum\x18\x04 \x01(\x01\"D\n\x0cLongsMessage\x12\r\n\x05\x63ount\x18\x01 \x01(\x03\x12\x0b\n\x03min\x18\x02 \x01(\x03\x12\x0b\n\x03max\x18\x03 \x01(\x03\x12\x0b\n\x03sum\x18\x04 \x01(\x03\";\n\x0fVarianceMessage\x12\r\n\x05\x63ount\x18\x01 \x01(\x03\x12\x0b\n\x03sum\x18\x02 \x01(\x01\x12\x0c\n\x04mean\x18\x03 \x01(\x01\"@\n\x1c\x46requentNumbersSketchMessage\x12\x0e\n\x06sketch\x18\x01 \x01(\x0c\x12\x10\n\x08lg_max_k\x18\x02 \x01(\x05\"@\n\x1c\x46requentItemsSketchMessageV0\x12\x0e\n\x06sketch\x18\x01 \x01(\x0c\x12\x10\n\x08lg_max_k\x18\x02 \x01(\x05\"\xf7\x01\n\x10NumbersMessageV0\x12\"\n\x08variance\x18\x01 \x01(\x0b\x32\x10.VarianceMessage\x12\"\n\x07\x64oubles\x18\x02 \x01(\x0b\x32\x0f.DoublesMessageH\x00\x12\x1e\n\x05longs\x18\x03 \x01(\x0b\x32\r.LongsMessageH\x00\x12\x11\n\thistogram\x18\x04 \x01(\x0c\x12\r\n\x05theta\x18\x05 \x01(\x0c\x12\x15\n\rcompact_theta\x18\x06 \x01(\x0c\x12\x37\n\x10\x66requent_numbers\x18\x07 \x01(\x0b\x32\x1d.FrequentNumbersSketchMessageB\t\n\x07numbers\"\xa0\x01\n\x0e\x43harPosMessage\x12\x11\n\tchar_list\x18\x01 \x01(\t\x12\x35\n\x0c\x63har_pos_map\x18\x02 \x03(\x0b\x32\x1f.CharPosMessage.CharPosMapEntry\x1a\x44\n\x0f\x43harPosMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12 \n\x05value\x18\x02 \x01(\x0b\x32\x11.NumbersMessageV0:\x02\x38\x01\"\xcd\x01\n\x10StringsMessageV0\x12\r\n\x05\x63ount\x18\x01 \x01(\x03\x12\r\n\x05theta\x18\x02 \x01(\x0c\x12\r\n\x05items\x18\x03 \x01(\x0c\x12\x15\n\rcompact_theta\x18\x04 \x01(\x0c\x12!\n\x06length\x18\x05 \x01(\x0b\x32\x11.NumbersMessageV0\x12\'\n\x0ctoken_length\x18\x06 \x01(\x0b\x32\x11.NumbersMessageV0\x12)\n\x10\x63har_pos_tracker\x18\x07 \x01(\x0b\x32\x0f.CharPosMessage\"\xa0\x01\n\x0fSchemaMessageV0\x12\x34\n\ntypeCounts\x18\x01 \x03(\x0b\x32 .SchemaMessageV0.TypeCountsEntry\x12$\n\rinferred_type\x18\x02 \x01(\x0b\x32\r.InferredType\x1a\x31\n\x0fTypeCountsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\"\xb7\x02\n\x0f\x43olumnMessageV0\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1d\n\x08\x63ounters\x18\x02 \x01(\x0b\x32\x0b.CountersV0\x12 \n\x06schema\x18\x03 \x01(\x0b\x32\x10.SchemaMessageV0\x12\"\n\x07numbers\x18\x04 \x01(\x0b\x32\x11.NumbersMessageV0\x12\"\n\x07strings\x18\x05 \x01(\x0b\x32\x11.StringsMessageV0\x12$\n\rinferred_type\x18\x06 \x01(\x0b\x32\r.InferredType\x12\x35\n\x0e\x66requent_items\x18\x07 \x01(\x0b\x32\x1d.FrequentItemsSketchMessageV0\x12\x30\n\x13\x63\x61rdinality_tracker\x18\x08 \x01(\x0b\x32\x13.HllSketchMessageV0\"\xda\x02\n\x13\x44\x61tasetPropertiesV0\x12\x1c\n\x14schema_major_version\x18\x01 \x01(\r\x12\x1c\n\x14schema_minor_version\x18\x02 \x01(\r\x12\x12\n\nsession_id\x18\x03 \x01(\t\x12\x19\n\x11session_timestamp\x18\x04 \x01(\x03\x12\x16\n\x0e\x64\x61ta_timestamp\x18\x05 \x01(\x03\x12,\n\x04tags\x18\x06 \x03(\x0b\x32\x1e.DatasetPropertiesV0.TagsEntry\x12\x34\n\x08metadata\x18\x07 \x03(\x0b\x32\".DatasetPropertiesV0.MetadataEntry\x1a+\n\tTagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a/\n\rMetadataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x8c\x01\n\x12ScoreMatrixMessage\x12\x0e\n\x06labels\x18\x01 \x03(\t\x12\x18\n\x10prediction_field\x18\x02 \x01(\t\x12\x14\n\x0ctarget_field\x18\x03 \x01(\t\x12\x13\n\x0bscore_field\x18\x04 \x01(\t\x12!\n\x06scores\x18\n \x03(\x0b\x32\x11.NumbersMessageV0\"\x94\x01\n\x18RegressionMetricsMessage\x12\x18\n\x10prediction_field\x18\x01 \x01(\t\x12\x14\n\x0ctarget_field\x18\x02 \x01(\t\x12\r\n\x05\x63ount\x18\x03 \x01(\x04\x12\x14\n\x0csum_abs_diff\x18\x04 \x01(\x01\x12\x10\n\x08sum_diff\x18\x05 \x01(\x01\x12\x11\n\tsum2_diff\x18\x06 \x01(\x01\"\x94\x01\n\x13ModelMetricsMessage\x12(\n\x0bscoreMatrix\x18\x01 \x01(\x0b\x32\x13.ScoreMatrixMessage\x12\x1d\n\tmodelType\x18\x02 \x01(\x0e\x32\n.ModelType\x12\x34\n\x11regressionMetrics\x18\x03 \x01(\x0b\x32\x19.RegressionMetricsMessage\"S\n\x13ModelProfileMessage\x12\x15\n\routput_fields\x18\x01 \x03(\t\x12%\n\x07metrics\x18\n \x01(\x0b\x32\x14.ModelMetricsMessage\"\xe8\x01\n\x17\x44\x61tasetProfileMessageV0\x12(\n\nproperties\x18\x01 \x01(\x0b\x32\x14.DatasetPropertiesV0\x12\x36\n\x07\x63olumns\x18\x02 \x03(\x0b\x32%.DatasetProfileMessageV0.ColumnsEntry\x12)\n\x0bmodeProfile\x18\n \x01(\x0b\x32\x14.ModelProfileMessage\x1a@\n\x0c\x43olumnsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1f\n\x05value\x18\x02 \x01(\x0b\x32\x10.ColumnMessageV0:\x02\x38\x01\"H\n\x13\x43olumnsChunkSegment\x12\x0e\n\x06marker\x18\x01 \x01(\t\x12!\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\x10.ColumnMessageV0\"R\n\x16\x44\x61tasetMetadataSegment\x12\x0e\n\x06marker\x18\x01 \x01(\t\x12(\n\nproperties\x18\x02 \x01(\x0b\x32\x14.DatasetPropertiesV0\"~\n\x0eMessageSegment\x12\x0e\n\x06marker\x18\x01 \x01(\t\x12+\n\x08metadata\x18\x02 \x01(\x0b\x32\x17.DatasetMetadataSegmentH\x00\x12\'\n\x07\x63olumns\x18\x03 \x01(\x0b\x32\x14.ColumnsChunkSegmentH\x00\x42\x06\n\x04item\"2\n\x12HllSketchMessageV0\x12\x0e\n\x06sketch\x18\x01 \x01(\x0c\x12\x0c\n\x04lg_k\x18\x02 \x01(\x05\"6\n\x16KllFloatsSketchMessage\x12\x0e\n\x06sketch\x18\x01 \x01(\x0c\x12\x0c\n\x04lg_k\x18\x02 \x01(\x05*U\n\tModelType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x12\n\x0e\x43LASSIFICATION\x10\x01\x12\x0e\n\nREGRESSION\x10\x02\x12\x0e\n\nEMBEDDINGS\x10\x03\x12\x07\n\x03NLP\x10\x04\x42)\n\x1b\x63om.whylogs.v0.core.messageB\x08MessagesP\x01\x62\x06proto3')

_MODELTYPE = DESCRIPTOR.enum_types_by_name['ModelType']
ModelType = enum_type_wrapper.EnumTypeWrapper(_MODELTYPE)
UNKNOWN = 0
CLASSIFICATION = 1
REGRESSION = 2
EMBEDDINGS = 3
NLP = 4


_COUNTERSV0 = DESCRIPTOR.message_types_by_name['CountersV0']
_INFERREDTYPE = DESCRIPTOR.message_types_by_name['InferredType']
_DOUBLESMESSAGE = DESCRIPTOR.message_types_by_name['DoublesMessage']
_LONGSMESSAGE = DESCRIPTOR.message_types_by_name['LongsMessage']
_VARIANCEMESSAGE = DESCRIPTOR.message_types_by_name['VarianceMessage']
_FREQUENTNUMBERSSKETCHMESSAGE = DESCRIPTOR.message_types_by_name['FrequentNumbersSketchMessage']
_FREQUENTITEMSSKETCHMESSAGEV0 = DESCRIPTOR.message_types_by_name['FrequentItemsSketchMessageV0']
_NUMBERSMESSAGEV0 = DESCRIPTOR.message_types_by_name['NumbersMessageV0']
_CHARPOSMESSAGE = DESCRIPTOR.message_types_by_name['CharPosMessage']
_CHARPOSMESSAGE_CHARPOSMAPENTRY = _CHARPOSMESSAGE.nested_types_by_name['CharPosMapEntry']
_STRINGSMESSAGEV0 = DESCRIPTOR.message_types_by_name['StringsMessageV0']
_SCHEMAMESSAGEV0 = DESCRIPTOR.message_types_by_name['SchemaMessageV0']
_SCHEMAMESSAGEV0_TYPECOUNTSENTRY = _SCHEMAMESSAGEV0.nested_types_by_name['TypeCountsEntry']
_COLUMNMESSAGEV0 = DESCRIPTOR.message_types_by_name['ColumnMessageV0']
_DATASETPROPERTIESV0 = DESCRIPTOR.message_types_by_name['DatasetPropertiesV0']
_DATASETPROPERTIESV0_TAGSENTRY = _DATASETPROPERTIESV0.nested_types_by_name['TagsEntry']
_DATASETPROPERTIESV0_METADATAENTRY = _DATASETPROPERTIESV0.nested_types_by_name['MetadataEntry']
_SCOREMATRIXMESSAGE = DESCRIPTOR.message_types_by_name['ScoreMatrixMessage']
_REGRESSIONMETRICSMESSAGE = DESCRIPTOR.message_types_by_name['RegressionMetricsMessage']
_MODELMETRICSMESSAGE = DESCRIPTOR.message_types_by_name['ModelMetricsMessage']
_MODELPROFILEMESSAGE = DESCRIPTOR.message_types_by_name['ModelProfileMessage']
_DATASETPROFILEMESSAGEV0 = DESCRIPTOR.message_types_by_name['DatasetProfileMessageV0']
_DATASETPROFILEMESSAGEV0_COLUMNSENTRY = _DATASETPROFILEMESSAGEV0.nested_types_by_name['ColumnsEntry']
_COLUMNSCHUNKSEGMENT = DESCRIPTOR.message_types_by_name['ColumnsChunkSegment']
_DATASETMETADATASEGMENT = DESCRIPTOR.message_types_by_name['DatasetMetadataSegment']
_MESSAGESEGMENT = DESCRIPTOR.message_types_by_name['MessageSegment']
_HLLSKETCHMESSAGEV0 = DESCRIPTOR.message_types_by_name['HllSketchMessageV0']
_KLLFLOATSSKETCHMESSAGE = DESCRIPTOR.message_types_by_name['KllFloatsSketchMessage']
_INFERREDTYPE_TYPE = _INFERREDTYPE.enum_types_by_name['Type']
CountersV0 = _reflection.GeneratedProtocolMessageType('CountersV0', (_message.Message,), {
  'DESCRIPTOR' : _COUNTERSV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:CountersV0)
  })
_sym_db.RegisterMessage(CountersV0)

InferredType = _reflection.GeneratedProtocolMessageType('InferredType', (_message.Message,), {
  'DESCRIPTOR' : _INFERREDTYPE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:InferredType)
  })
_sym_db.RegisterMessage(InferredType)

DoublesMessage = _reflection.GeneratedProtocolMessageType('DoublesMessage', (_message.Message,), {
  'DESCRIPTOR' : _DOUBLESMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:DoublesMessage)
  })
_sym_db.RegisterMessage(DoublesMessage)

LongsMessage = _reflection.GeneratedProtocolMessageType('LongsMessage', (_message.Message,), {
  'DESCRIPTOR' : _LONGSMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:LongsMessage)
  })
_sym_db.RegisterMessage(LongsMessage)

VarianceMessage = _reflection.GeneratedProtocolMessageType('VarianceMessage', (_message.Message,), {
  'DESCRIPTOR' : _VARIANCEMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:VarianceMessage)
  })
_sym_db.RegisterMessage(VarianceMessage)

FrequentNumbersSketchMessage = _reflection.GeneratedProtocolMessageType('FrequentNumbersSketchMessage', (_message.Message,), {
  'DESCRIPTOR' : _FREQUENTNUMBERSSKETCHMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:FrequentNumbersSketchMessage)
  })
_sym_db.RegisterMessage(FrequentNumbersSketchMessage)

FrequentItemsSketchMessageV0 = _reflection.GeneratedProtocolMessageType('FrequentItemsSketchMessageV0', (_message.Message,), {
  'DESCRIPTOR' : _FREQUENTITEMSSKETCHMESSAGEV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:FrequentItemsSketchMessageV0)
  })
_sym_db.RegisterMessage(FrequentItemsSketchMessageV0)

NumbersMessageV0 = _reflection.GeneratedProtocolMessageType('NumbersMessageV0', (_message.Message,), {
  'DESCRIPTOR' : _NUMBERSMESSAGEV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:NumbersMessageV0)
  })
_sym_db.RegisterMessage(NumbersMessageV0)

CharPosMessage = _reflection.GeneratedProtocolMessageType('CharPosMessage', (_message.Message,), {

  'CharPosMapEntry' : _reflection.GeneratedProtocolMessageType('CharPosMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _CHARPOSMESSAGE_CHARPOSMAPENTRY,
    '__module__' : 'v0_messages_pb2'
    # @@protoc_insertion_point(class_scope:CharPosMessage.CharPosMapEntry)
    })
  ,
  'DESCRIPTOR' : _CHARPOSMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:CharPosMessage)
  })
_sym_db.RegisterMessage(CharPosMessage)
_sym_db.RegisterMessage(CharPosMessage.CharPosMapEntry)

StringsMessageV0 = _reflection.GeneratedProtocolMessageType('StringsMessageV0', (_message.Message,), {
  'DESCRIPTOR' : _STRINGSMESSAGEV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:StringsMessageV0)
  })
_sym_db.RegisterMessage(StringsMessageV0)

SchemaMessageV0 = _reflection.GeneratedProtocolMessageType('SchemaMessageV0', (_message.Message,), {

  'TypeCountsEntry' : _reflection.GeneratedProtocolMessageType('TypeCountsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SCHEMAMESSAGEV0_TYPECOUNTSENTRY,
    '__module__' : 'v0_messages_pb2'
    # @@protoc_insertion_point(class_scope:SchemaMessageV0.TypeCountsEntry)
    })
  ,
  'DESCRIPTOR' : _SCHEMAMESSAGEV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:SchemaMessageV0)
  })
_sym_db.RegisterMessage(SchemaMessageV0)
_sym_db.RegisterMessage(SchemaMessageV0.TypeCountsEntry)

ColumnMessageV0 = _reflection.GeneratedProtocolMessageType('ColumnMessageV0', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNMESSAGEV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:ColumnMessageV0)
  })
_sym_db.RegisterMessage(ColumnMessageV0)

DatasetPropertiesV0 = _reflection.GeneratedProtocolMessageType('DatasetPropertiesV0', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETPROPERTIESV0_TAGSENTRY,
    '__module__' : 'v0_messages_pb2'
    # @@protoc_insertion_point(class_scope:DatasetPropertiesV0.TagsEntry)
    })
  ,

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETPROPERTIESV0_METADATAENTRY,
    '__module__' : 'v0_messages_pb2'
    # @@protoc_insertion_point(class_scope:DatasetPropertiesV0.MetadataEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETPROPERTIESV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:DatasetPropertiesV0)
  })
_sym_db.RegisterMessage(DatasetPropertiesV0)
_sym_db.RegisterMessage(DatasetPropertiesV0.TagsEntry)
_sym_db.RegisterMessage(DatasetPropertiesV0.MetadataEntry)

ScoreMatrixMessage = _reflection.GeneratedProtocolMessageType('ScoreMatrixMessage', (_message.Message,), {
  'DESCRIPTOR' : _SCOREMATRIXMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:ScoreMatrixMessage)
  })
_sym_db.RegisterMessage(ScoreMatrixMessage)

RegressionMetricsMessage = _reflection.GeneratedProtocolMessageType('RegressionMetricsMessage', (_message.Message,), {
  'DESCRIPTOR' : _REGRESSIONMETRICSMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:RegressionMetricsMessage)
  })
_sym_db.RegisterMessage(RegressionMetricsMessage)

ModelMetricsMessage = _reflection.GeneratedProtocolMessageType('ModelMetricsMessage', (_message.Message,), {
  'DESCRIPTOR' : _MODELMETRICSMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:ModelMetricsMessage)
  })
_sym_db.RegisterMessage(ModelMetricsMessage)

ModelProfileMessage = _reflection.GeneratedProtocolMessageType('ModelProfileMessage', (_message.Message,), {
  'DESCRIPTOR' : _MODELPROFILEMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:ModelProfileMessage)
  })
_sym_db.RegisterMessage(ModelProfileMessage)

DatasetProfileMessageV0 = _reflection.GeneratedProtocolMessageType('DatasetProfileMessageV0', (_message.Message,), {

  'ColumnsEntry' : _reflection.GeneratedProtocolMessageType('ColumnsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETPROFILEMESSAGEV0_COLUMNSENTRY,
    '__module__' : 'v0_messages_pb2'
    # @@protoc_insertion_point(class_scope:DatasetProfileMessageV0.ColumnsEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETPROFILEMESSAGEV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:DatasetProfileMessageV0)
  })
_sym_db.RegisterMessage(DatasetProfileMessageV0)
_sym_db.RegisterMessage(DatasetProfileMessageV0.ColumnsEntry)

ColumnsChunkSegment = _reflection.GeneratedProtocolMessageType('ColumnsChunkSegment', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNSCHUNKSEGMENT,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:ColumnsChunkSegment)
  })
_sym_db.RegisterMessage(ColumnsChunkSegment)

DatasetMetadataSegment = _reflection.GeneratedProtocolMessageType('DatasetMetadataSegment', (_message.Message,), {
  'DESCRIPTOR' : _DATASETMETADATASEGMENT,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:DatasetMetadataSegment)
  })
_sym_db.RegisterMessage(DatasetMetadataSegment)

MessageSegment = _reflection.GeneratedProtocolMessageType('MessageSegment', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGESEGMENT,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:MessageSegment)
  })
_sym_db.RegisterMessage(MessageSegment)

HllSketchMessageV0 = _reflection.GeneratedProtocolMessageType('HllSketchMessageV0', (_message.Message,), {
  'DESCRIPTOR' : _HLLSKETCHMESSAGEV0,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:HllSketchMessageV0)
  })
_sym_db.RegisterMessage(HllSketchMessageV0)

KllFloatsSketchMessage = _reflection.GeneratedProtocolMessageType('KllFloatsSketchMessage', (_message.Message,), {
  'DESCRIPTOR' : _KLLFLOATSSKETCHMESSAGE,
  '__module__' : 'v0_messages_pb2'
  # @@protoc_insertion_point(class_scope:KllFloatsSketchMessage)
  })
_sym_db.RegisterMessage(KllFloatsSketchMessage)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\033com.whylogs.v0.core.messageB\010MessagesP\001'
  _COUNTERSV0.fields_by_name['null_count']._options = None
  _COUNTERSV0.fields_by_name['null_count']._serialized_options = b'\030\001'
  _CHARPOSMESSAGE_CHARPOSMAPENTRY._options = None
  _CHARPOSMESSAGE_CHARPOSMAPENTRY._serialized_options = b'8\001'
  _SCHEMAMESSAGEV0_TYPECOUNTSENTRY._options = None
  _SCHEMAMESSAGEV0_TYPECOUNTSENTRY._serialized_options = b'8\001'
  _DATASETPROPERTIESV0_TAGSENTRY._options = None
  _DATASETPROPERTIESV0_TAGSENTRY._serialized_options = b'8\001'
  _DATASETPROPERTIESV0_METADATAENTRY._options = None
  _DATASETPROPERTIESV0_METADATAENTRY._serialized_options = b'8\001'
  _DATASETPROFILEMESSAGEV0_COLUMNSENTRY._options = None
  _DATASETPROFILEMESSAGEV0_COLUMNSENTRY._serialized_options = b'8\001'
  _MODELTYPE._serialized_start=3278
  _MODELTYPE._serialized_end=3363
  _COUNTERSV0._serialized_start=54
  _COUNTERSV0._serialized_end=183
  _INFERREDTYPE._serialized_start=186
  _INFERREDTYPE._serialized_end=335
  _INFERREDTYPE_TYPE._serialized_start=251
  _INFERREDTYPE_TYPE._serialized_end=335
  _DOUBLESMESSAGE._serialized_start=337
  _DOUBLESMESSAGE._serialized_end=407
  _LONGSMESSAGE._serialized_start=409
  _LONGSMESSAGE._serialized_end=477
  _VARIANCEMESSAGE._serialized_start=479
  _VARIANCEMESSAGE._serialized_end=538
  _FREQUENTNUMBERSSKETCHMESSAGE._serialized_start=540
  _FREQUENTNUMBERSSKETCHMESSAGE._serialized_end=604
  _FREQUENTITEMSSKETCHMESSAGEV0._serialized_start=606
  _FREQUENTITEMSSKETCHMESSAGEV0._serialized_end=670
  _NUMBERSMESSAGEV0._serialized_start=673
  _NUMBERSMESSAGEV0._serialized_end=920
  _CHARPOSMESSAGE._serialized_start=923
  _CHARPOSMESSAGE._serialized_end=1083
  _CHARPOSMESSAGE_CHARPOSMAPENTRY._serialized_start=1015
  _CHARPOSMESSAGE_CHARPOSMAPENTRY._serialized_end=1083
  _STRINGSMESSAGEV0._serialized_start=1086
  _STRINGSMESSAGEV0._serialized_end=1291
  _SCHEMAMESSAGEV0._serialized_start=1294
  _SCHEMAMESSAGEV0._serialized_end=1454
  _SCHEMAMESSAGEV0_TYPECOUNTSENTRY._serialized_start=1405
  _SCHEMAMESSAGEV0_TYPECOUNTSENTRY._serialized_end=1454
  _COLUMNMESSAGEV0._serialized_start=1457
  _COLUMNMESSAGEV0._serialized_end=1768
  _DATASETPROPERTIESV0._serialized_start=1771
  _DATASETPROPERTIESV0._serialized_end=2117
  _DATASETPROPERTIESV0_TAGSENTRY._serialized_start=2025
  _DATASETPROPERTIESV0_TAGSENTRY._serialized_end=2068
  _DATASETPROPERTIESV0_METADATAENTRY._serialized_start=2070
  _DATASETPROPERTIESV0_METADATAENTRY._serialized_end=2117
  _SCOREMATRIXMESSAGE._serialized_start=2120
  _SCOREMATRIXMESSAGE._serialized_end=2260
  _REGRESSIONMETRICSMESSAGE._serialized_start=2263
  _REGRESSIONMETRICSMESSAGE._serialized_end=2411
  _MODELMETRICSMESSAGE._serialized_start=2414
  _MODELMETRICSMESSAGE._serialized_end=2562
  _MODELPROFILEMESSAGE._serialized_start=2564
  _MODELPROFILEMESSAGE._serialized_end=2647
  _DATASETPROFILEMESSAGEV0._serialized_start=2650
  _DATASETPROFILEMESSAGEV0._serialized_end=2882
  _DATASETPROFILEMESSAGEV0_COLUMNSENTRY._serialized_start=2818
  _DATASETPROFILEMESSAGEV0_COLUMNSENTRY._serialized_end=2882
  _COLUMNSCHUNKSEGMENT._serialized_start=2884
  _COLUMNSCHUNKSEGMENT._serialized_end=2956
  _DATASETMETADATASEGMENT._serialized_start=2958
  _DATASETMETADATASEGMENT._serialized_end=3040
  _MESSAGESEGMENT._serialized_start=3042
  _MESSAGESEGMENT._serialized_end=3168
  _HLLSKETCHMESSAGEV0._serialized_start=3170
  _HLLSKETCHMESSAGEV0._serialized_end=3220
  _KLLFLOATSSKETCHMESSAGE._serialized_start=3222
  _KLLFLOATSSKETCHMESSAGE._serialized_end=3276
# @@protoc_insertion_point(module_scope)



import hashlib
import inspect
import logging
from dataclasses import dataclass, field
from typing import Callable, List, Mapping, Optional

from whylogs.core.projectors import FieldProjector

logger = logging.getLogger(__name__)

MAX_SEGMENTS_ERROR_THRESHOLD = 2000  # if profiles are up to 500KiB each, then 500KiB * 2000 = 1 GiB
MAX_SEGMENTS_WARNING_THRESHOLD = 200


@dataclass(frozen=False)
class ColumnMapperFunction:
    col_names: Optional[List[str]] = None
    map: Optional[Callable] = None
    field_projector: FieldProjector = field(init=False)
    id: str = field(init=False)

    def __post_init__(self):
        self.field_projector = FieldProjector(col_names=self.col_names)
        mapper_string = inspect.getsource(self.map) if self.map else ""
        column_string = ",".join(sorted(self.col_names))
        segment_hash = hashlib.sha512(bytes(column_string + mapper_string, encoding="utf8"))
        self.id = segment_hash.hexdigest()


@dataclass
class SegmentFilter:
    filter_function: Optional[Callable] = None
    query_string: Optional[str] = None


@dataclass(frozen=False)
class SegmentationPartition:
    name: str
    mapper: Optional[ColumnMapperFunction] = None
    id: str = field(init=False)
    filter: Optional[SegmentFilter] = None

    def __post_init__(self):
        filter_string = self.filter_id
        mapper_string = self.mapper.id if self.mapper else ""
        segment_hash = hashlib.sha512(bytes(filter_string + mapper_string, encoding="utf8"))
        self.id = segment_hash.hexdigest()

    @property
    def simple(self) -> bool:
        return self.mapper is None or self.mapper.map is None

    @property
    def filter_id(self) -> str:
        if self.filter:
            if self.filter.filter_function:
                return inspect.getsource(self.filter.filter_function)
            if self.filter.query_string:
                return self.filter.query_string
        return ""


def segment_on_column(column_name: str) -> Mapping[str, SegmentationPartition]:
    return {column_name: SegmentationPartition(name=column_name, mapper=ColumnMapperFunction(col_names=[column_name]))}



from logging import getLogger
from typing import Dict, List, Optional, Tuple, Union

import whylogs_sketching as ds  # type: ignore

from whylogs.core.metrics.metric_components import FractionalComponent, KllComponent
from whylogs.core.metrics.metrics import DistributionMetric, MetricConfig
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.proto.v0 import (
    DoublesMessage,
    NumbersMessageV0,
    ScoreMatrixMessage,
    VarianceMessage,
)

MODEL_METRICS_MAX_LABELS = 256
MODEL_METRICS_LABEL_SIZE_WARNING_THRESHOLD = 64
EMPTY_KLL: bytes = ds.kll_doubles_sketch(k=128).serialize()
_empty_theta_union = ds.theta_union()
_empty_theta_union.update(ds.update_theta_sketch())
EMPTY_THETA: bytes = _empty_theta_union.get_result().serialize()

_logger = getLogger("whylogs")


def _encode_to_integers(values, uniques):
    table = {val: i for i, val in enumerate(uniques)}
    for v in values:
        if v not in uniques:
            raise ValueError("Can not encode values not in unique set")
    return [table[v] for v in values]


class ConfusionMatrix:
    """

    Confusion Matrix Class to hold labels and matrix data.

    Attributes:
        labels: list of labels in a sorted order
    """

    def __init__(
        self,
        labels: List[Union[str, int, bool, float]] = None,
    ):
        if labels:
            labels_size = len(labels)
            if labels_size > MODEL_METRICS_LABEL_SIZE_WARNING_THRESHOLD:
                _logger.warning(
                    f"The initialized confusion matrix has {labels_size} labels and the resulting"
                    " confusion matrix will be larger than is recommended with whylogs current"
                    " representation of the model metric for a confusion matrix of this size."
                )
            if labels_size > MODEL_METRICS_MAX_LABELS:
                raise ValueError(
                    f"The initialized confusion matrix has {labels_size} labels and the resulting"
                    " confusion matrix will be larger than is supported by whylogs current"
                    " representation of the model metric for a confusion matrix of this size,"
                    " selectively log the most important labels or configure the threshold of"
                    " {MODEL_METRICS_MAX_LABELS} higher by setting MODEL_METRICS_MAX_LABELS."
                )

            self.labels = sorted(labels)
        else:
            self.labels = list()

        self.confusion_matrix: Dict[Tuple[int, int], DistributionMetric] = dict()
        self.default_config = MetricConfig()

    def add(
        self,
        predictions: List[Union[str, int, bool, float]],
        targets: List[Union[str, int, bool, float]],
        scores: Optional[List[float]],
    ):
        """
        Function adds predictions and targets to confusion matrix with scores.

        Args:
            predictions (List[Union[str, int, bool]]):
            targets (List[Union[str, int, bool]]):
            scores (List[float]):

        Raises:
            NotImplementedError: in case targets do not fall into binary or
            multiclass suport
            ValueError: incase missing validation or predictions
        """
        if not isinstance(targets, list):
            targets = [targets]
        if not isinstance(predictions, list):
            predictions = [predictions]

        if scores is None:
            scores = [1.0 for _ in range(len(targets))]

        if len(targets) != len(predictions):
            raise ValueError("both targets and predictions need to have the same length")

        targets_indx = _encode_to_integers(targets, self.labels)
        prediction_indx = _encode_to_integers(predictions, self.labels)

        for ind in range(len(predictions)):
            entry_key = prediction_indx[ind], targets_indx[ind]
            if entry_key not in self.confusion_matrix:
                self.confusion_matrix[entry_key] = DistributionMetric.zero(self.default_config)
            data = PreprocessedColumn.apply([scores[ind]])
            self.confusion_matrix[entry_key].columnar_update(data)

    def merge(self, other_cm):
        """
        Merge two seperate confusion matrix which may or may not overlap in labels.

        Args:
              other_cm (Optional[ConfusionMatrix]): confusion_matrix to merge with self
        Returns:
              ConfusionMatrix: merged confusion_matrix
        """
        # TODO: always return new objects
        if other_cm is None:
            return self
        if self.labels is None or self.labels == []:
            return other_cm
        if other_cm.labels is None or other_cm.labels == []:
            return self

        # the union of the labels potentially creates a new encoding
        labels = list(set(self.labels + other_cm.labels))

        conf_matrix = ConfusionMatrix(labels)

        conf_matrix = _merge_CM(self, conf_matrix)
        conf_matrix = _merge_CM(other_cm, conf_matrix)

        return conf_matrix

    @staticmethod
    def _dist_to_numbers(dist: Optional[DistributionMetric]) -> NumbersMessageV0:

        variance_message = VarianceMessage()

        if dist is None or dist.kll.value.is_empty():
            return NumbersMessageV0(histogram=EMPTY_KLL, compact_theta=EMPTY_THETA, variance=variance_message)

        variance_message = VarianceMessage(count=dist.n, sum=dist.m2.value, mean=dist.mean.value)
        return NumbersMessageV0(
            histogram=dist.kll.value.serialize(),
            compact_theta=EMPTY_THETA,
            variance=variance_message,
            doubles=DoublesMessage(count=dist.n),
        )

    @staticmethod
    def _numbers_to_dist(numbers: NumbersMessageV0) -> DistributionMetric:
        doubles_sk = ds.kll_doubles_sketch.deserialize(numbers.histogram)
        return DistributionMetric(
            kll=KllComponent(doubles_sk),
            mean=FractionalComponent(numbers.variance.mean),
            m2=FractionalComponent(numbers.variance.sum),
        )

    def to_protobuf(
        self,
    ):
        """
        Convert to protobuf

        Returns:
            TYPE: Description
        """
        size = len(self.labels)
        if size == 0:
            return None
        confusion_matrix_entries: List[NumbersMessageV0] = []
        for i in range(size):
            for j in range(size):
                entry_key = i, j
                entry = self.confusion_matrix.get(entry_key)
                numbers_message = ConfusionMatrix._dist_to_numbers(entry)
                confusion_matrix_entries.append(numbers_message)

        return ScoreMatrixMessage(
            labels=[str(i) for i in self.labels],
            scores=confusion_matrix_entries,
        )

    @classmethod
    def from_protobuf(
        cls,
        message: ScoreMatrixMessage,
    ):
        if message is None or message.ByteSize() == 0:
            return None
        labels = message.labels
        num_labels = len(labels)
        matrix = dict()
        for i in range(num_labels):
            for j in range(num_labels):
                index = i * num_labels + j
                entry_key = i, j
                entry = message.scores[index]
                matrix[entry_key] = ConfusionMatrix._numbers_to_dist(entry)

        cm_instance = ConfusionMatrix(
            labels=labels,
        )
        cm_instance.confusion_matrix = matrix

        return cm_instance


def _merge_CM(old_conf_matrix: ConfusionMatrix, new_conf_matrix: ConfusionMatrix):
    """
    Merges two confusion_matrix since distinc or overlaping labels

    Args:
        old_conf_matrix (ConfusionMatrix)
        new_conf_matrix (ConfusionMatrix): Will be overridden
    """
    new_indxes = _encode_to_integers(old_conf_matrix.labels, new_conf_matrix.labels)
    old_indxes = _encode_to_integers(old_conf_matrix.labels, old_conf_matrix.labels)

    res_conf_matrix = ConfusionMatrix(new_conf_matrix.labels)

    res_conf_matrix.confusion_matrix = new_conf_matrix.confusion_matrix

    for old_row_idx, each_row_indx in enumerate(new_indxes):
        for old_column_idx, each_column_inx in enumerate(new_indxes):
            new_entry = new_conf_matrix.confusion_matrix.get((each_row_indx, each_column_inx))
            old_entry = old_conf_matrix.confusion_matrix.get((old_indxes[old_row_idx], old_indxes[old_column_idx]))
            res_conf_matrix.confusion_matrix[each_row_indx, each_column_inx] = (
                new_entry.merge(old_entry) if new_entry else old_entry
            )

    return res_conf_matrix



from typing import Optional, Tuple, Union

from typing_extensions import TypedDict

from whylogs.core.metrics import ColumnCountsMetric, DistributionMetric
from whylogs.core.view.column_profile_view import ColumnProfileView


class DescriptiveStatistics(TypedDict):
    """Collection of descriptive statistics, such as standard deviation,
    coefficient of variation, sum, variance and mean."""

    stddev: Optional[float]
    coefficient_of_variation: Optional[float]
    sum: Optional[float]
    variance: Optional[float]
    mean: Optional[float]


def _get_count_metrics_from_column_view(
    column_view: ColumnProfileView,
) -> Union[Tuple[None, None, None], Tuple[float, float, float]]:
    column_counts_metric: ColumnCountsMetric = column_view.get_metric("counts")  # type: ignore
    if column_counts_metric is None:
        return None, None, None
    count_n = column_counts_metric.n.value
    count_missing = column_counts_metric.null.value
    total_count = count_n - count_missing
    return total_count, count_n, count_missing


def _get_dist_metrics_from_column_view(
    column_view: ColumnProfileView,
) -> Union[Tuple[None, None, None], Tuple[float, float, float]]:
    distribution_metric: Optional[DistributionMetric] = column_view.get_metric("distribution")
    if distribution_metric is None:
        return None, None, None
    stddev = distribution_metric.stddev
    mean = distribution_metric.mean.value
    variance = distribution_metric.variance
    return stddev, mean, variance


def _calculate_descriptive_statistics(
    column_view: Union[ColumnProfileView, None]
) -> Union[None, DescriptiveStatistics]:
    if column_view is None:
        return None

    stddev, mean, variance = _get_dist_metrics_from_column_view(column_view=column_view)
    _, count_n, count_missing = _get_count_metrics_from_column_view(column_view=column_view)

    if count_n is not None and count_missing is not None and mean is not None:
        sum_ = (count_n - count_missing) * mean
    else:
        sum_ = None

    if stddev is not None and mean:
        coefficient_of_variation = stddev / mean
    else:
        coefficient_of_variation = None

    descriptive_statistics: DescriptiveStatistics = {
        "stddev": stddev,
        "coefficient_of_variation": coefficient_of_variation,
        "sum": sum_,
        "variance": variance,
        "mean": mean,
    }
    return descriptive_statistics



from dataclasses import dataclass
from enum import Enum


@dataclass
class PageSpec:
    html: str
    height: str


class PageSpecEnum(Enum):
    PROFILE_SUMMARY = PageSpec(html="index-hbs-cdn-all-in-jupyter-profile-summary.html", height="1000px")
    SUMMARY_REPORT = PageSpec(html="index-hbs-cdn-all-in-for-jupyter-notebook.html", height="1000px")
    DOUBLE_HISTOGRAM = PageSpec(html="index-hbs-cdn-all-in-jupyter-distribution-chart.html", height="300px")
    DISTRIBUTION_CHART = PageSpec(html="index-hbs-cdn-all-in-jupyter-bar-chart.html", height="277px")
    DIFFERENCED_CHART = PageSpec(html="index-hbs-cdn-all-in-jupyter-differenced-chart.html", height="277px")
    FEATURE_STATISTICS = PageSpec(html="index-hbs-cdn-all-in-jupyter-feature-summary-statistics.html", height="650px")
    CONSTRAINTS_REPORT = PageSpec(html="index-hbs-cdn-all-in-jupyter-constraints-report.html", height="750px")



from enum import Enum

from whylogs.core.metrics.column_metrics import ColumnCountsMetric, TypeCountersMetric
from whylogs.core.metrics.condition_count_metric import ConditionCountMetric
from whylogs.core.metrics.metrics import (
    CardinalityMetric,
    DistributionMetric,
    FrequentItemsMetric,
    IntsMetric,
    Metric,
    MetricConfig,
)
from whylogs.core.metrics.unicode_range import UnicodeRangeMetric


class StandardMetric(Enum):
    types = TypeCountersMetric
    distribution = DistributionMetric
    counts = ColumnCountsMetric
    ints = IntsMetric
    cardinality = CardinalityMetric
    frequent_items = FrequentItemsMetric
    unicode_range = UnicodeRangeMetric
    condition_count = ConditionCountMetric

    def __init__(self, clz: Metric):
        self._clz = clz

    def zero(self, config: MetricConfig) -> Metric:  # type: ignore
        return self._clz.zero(config)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.struct_pb2
import typing
import typing_extensions
import v0_messages_pb2
import v0_summaries_pb2

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _Op:
    ValueType = typing.NewType('ValueType', builtins.int)
    V: typing_extensions.TypeAlias = ValueType
class _OpEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Op.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    unused: _Op.ValueType  # 0
    LT: _Op.ValueType  # 1
    LE: _Op.ValueType  # 2
    EQ: _Op.ValueType  # 3
    NE: _Op.ValueType  # 4
    GE: _Op.ValueType  # 5
    GT: _Op.ValueType  # 6
    MATCH: _Op.ValueType  # 7
    NOMATCH: _Op.ValueType  # 8
    BTWN: _Op.ValueType  # 9
    IN_SET: _Op.ValueType  # 10
    CONTAIN_SET: _Op.ValueType  # 11
    EQ_SET: _Op.ValueType  # 12
    APPLY_FUNC: _Op.ValueType  # 13
    IN: _Op.ValueType  # 14
    CONTAIN: _Op.ValueType  # 15
    NOT_IN: _Op.ValueType  # 16
    SUM: _Op.ValueType  # 17
class Op(_Op, metaclass=_OpEnumTypeWrapper):
    """constraints specify one of the following binary boolean relationships."""
    pass

unused: Op.ValueType  # 0
LT: Op.ValueType  # 1
LE: Op.ValueType  # 2
EQ: Op.ValueType  # 3
NE: Op.ValueType  # 4
GE: Op.ValueType  # 5
GT: Op.ValueType  # 6
MATCH: Op.ValueType  # 7
NOMATCH: Op.ValueType  # 8
BTWN: Op.ValueType  # 9
IN_SET: Op.ValueType  # 10
CONTAIN_SET: Op.ValueType  # 11
EQ_SET: Op.ValueType  # 12
APPLY_FUNC: Op.ValueType  # 13
IN: Op.ValueType  # 14
CONTAIN: Op.ValueType  # 15
NOT_IN: Op.ValueType  # 16
SUM: Op.ValueType  # 17
global___Op = Op


class SummaryConstraintMsg(google.protobuf.message.Message):
    """Summary constraints specify a relationship between a summary field and a literal value,
    or between two summary fields.
    e.g.     'min' < 6
    'std_dev' < 2.17
    'min' > 'avg'
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    FIRST_FIELD_FIELD_NUMBER: builtins.int
    SECOND_FIELD_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    BETWEEN_FIELD_NUMBER: builtins.int
    REFERENCE_SET_FIELD_NUMBER: builtins.int
    VALUE_STR_FIELD_NUMBER: builtins.int
    OP_FIELD_NUMBER: builtins.int
    VERBOSE_FIELD_NUMBER: builtins.int
    QUANTILE_VALUE_FIELD_NUMBER: builtins.int
    CONTINUOUS_DISTRIBUTION_FIELD_NUMBER: builtins.int
    DISCRETE_DISTRIBUTION_FIELD_NUMBER: builtins.int
    name: typing.Text
    first_field: typing.Text
    second_field: typing.Text
    value: builtins.float
    @property
    def between(self) -> global___SummaryBetweenConstraintMsg: ...
    @property
    def reference_set(self) -> google.protobuf.struct_pb2.ListValue: ...
    value_str: typing.Text
    op: global___Op.ValueType
    verbose: builtins.bool
    quantile_value: builtins.float
    @property
    def continuous_distribution(self) -> global___ReferenceDistributionContinuousMessage: ...
    @property
    def discrete_distribution(self) -> global___ReferenceDistributionDiscreteMessage: ...
    def __init__(self,
        *,
        name: typing.Text = ...,
        first_field: typing.Text = ...,
        second_field: typing.Text = ...,
        value: builtins.float = ...,
        between: typing.Optional[global___SummaryBetweenConstraintMsg] = ...,
        reference_set: typing.Optional[google.protobuf.struct_pb2.ListValue] = ...,
        value_str: typing.Text = ...,
        op: global___Op.ValueType = ...,
        verbose: builtins.bool = ...,
        quantile_value: builtins.float = ...,
        continuous_distribution: typing.Optional[global___ReferenceDistributionContinuousMessage] = ...,
        discrete_distribution: typing.Optional[global___ReferenceDistributionDiscreteMessage] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["between",b"between","continuous_distribution",b"continuous_distribution","discrete_distribution",b"discrete_distribution","reference_distribution",b"reference_distribution","reference_set",b"reference_set","second",b"second","second_field",b"second_field","value",b"value","value_str",b"value_str"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["between",b"between","continuous_distribution",b"continuous_distribution","discrete_distribution",b"discrete_distribution","first_field",b"first_field","name",b"name","op",b"op","quantile_value",b"quantile_value","reference_distribution",b"reference_distribution","reference_set",b"reference_set","second",b"second","second_field",b"second_field","value",b"value","value_str",b"value_str","verbose",b"verbose"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["reference_distribution",b"reference_distribution"]) -> typing.Optional[typing_extensions.Literal["continuous_distribution","discrete_distribution"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["second",b"second"]) -> typing.Optional[typing_extensions.Literal["second_field","value","between","reference_set","value_str"]]: ...
global___SummaryConstraintMsg = SummaryConstraintMsg

class ReferenceDistributionContinuousMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SKETCH_FIELD_NUMBER: builtins.int
    @property
    def sketch(self) -> v0_messages_pb2.KllFloatsSketchMessage: ...
    def __init__(self,
        *,
        sketch: typing.Optional[v0_messages_pb2.KllFloatsSketchMessage] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["sketch",b"sketch"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["sketch",b"sketch"]) -> None: ...
global___ReferenceDistributionContinuousMessage = ReferenceDistributionContinuousMessage

class ReferenceDistributionDiscreteMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FREQUENT_ITEMS_FIELD_NUMBER: builtins.int
    UNIQUE_COUNT_FIELD_NUMBER: builtins.int
    TOTAL_COUNT_FIELD_NUMBER: builtins.int
    @property
    def frequent_items(self) -> v0_summaries_pb2.FrequentItemsSummary: ...
    @property
    def unique_count(self) -> v0_summaries_pb2.UniqueCountSummary: ...
    total_count: builtins.float
    def __init__(self,
        *,
        frequent_items: typing.Optional[v0_summaries_pb2.FrequentItemsSummary] = ...,
        unique_count: typing.Optional[v0_summaries_pb2.UniqueCountSummary] = ...,
        total_count: builtins.float = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["frequent_items",b"frequent_items","unique_count",b"unique_count"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["frequent_items",b"frequent_items","total_count",b"total_count","unique_count",b"unique_count"]) -> None: ...
global___ReferenceDistributionDiscreteMessage = ReferenceDistributionDiscreteMessage

class SummaryBetweenConstraintMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SECOND_FIELD_FIELD_NUMBER: builtins.int
    LOWER_VALUE_FIELD_NUMBER: builtins.int
    THIRD_FIELD_FIELD_NUMBER: builtins.int
    UPPER_VALUE_FIELD_NUMBER: builtins.int
    second_field: typing.Text
    lower_value: builtins.float
    third_field: typing.Text
    upper_value: builtins.float
    def __init__(self,
        *,
        second_field: typing.Text = ...,
        lower_value: builtins.float = ...,
        third_field: typing.Text = ...,
        upper_value: builtins.float = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["lower",b"lower","lower_value",b"lower_value","second_field",b"second_field","third_field",b"third_field","upper",b"upper","upper_value",b"upper_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["lower",b"lower","lower_value",b"lower_value","second_field",b"second_field","third_field",b"third_field","upper",b"upper","upper_value",b"upper_value"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["lower",b"lower"]) -> typing.Optional[typing_extensions.Literal["second_field","lower_value"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["upper",b"upper"]) -> typing.Optional[typing_extensions.Literal["third_field","upper_value"]]: ...
global___SummaryBetweenConstraintMsg = SummaryBetweenConstraintMsg

class ApplyFunctionMsg(google.protobuf.message.Message):
    """ValueConstraints express a binary boolean relationship between an implied numeric value and a literal, or between a string value and a regex pattern.
    These are applied to every incoming value that is processed by whylogs.
    """
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    FUNCTION_FIELD_NUMBER: builtins.int
    REFERENCE_VALUE_FIELD_NUMBER: builtins.int
    function: typing.Text
    reference_value: typing.Text
    def __init__(self,
        *,
        function: typing.Text = ...,
        reference_value: typing.Text = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["function",b"function","reference_value",b"reference_value"]) -> None: ...
global___ApplyFunctionMsg = ApplyFunctionMsg

class ValueConstraintMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    REGEX_PATTERN_FIELD_NUMBER: builtins.int
    VALUE_SET_FIELD_NUMBER: builtins.int
    FUNCTION_FIELD_NUMBER: builtins.int
    OP_FIELD_NUMBER: builtins.int
    VERBOSE_FIELD_NUMBER: builtins.int
    TOTAL_FIELD_NUMBER: builtins.int
    FAILURES_FIELD_NUMBER: builtins.int
    name: typing.Text
    value: builtins.float
    regex_pattern: typing.Text
    @property
    def value_set(self) -> google.protobuf.struct_pb2.ListValue: ...
    @property
    def function(self) -> global___ApplyFunctionMsg: ...
    op: global___Op.ValueType
    verbose: builtins.bool
    total: builtins.int
    failures: builtins.int
    def __init__(self,
        *,
        name: typing.Text = ...,
        value: builtins.float = ...,
        regex_pattern: typing.Text = ...,
        value_set: typing.Optional[google.protobuf.struct_pb2.ListValue] = ...,
        function: typing.Optional[global___ApplyFunctionMsg] = ...,
        op: global___Op.ValueType = ...,
        verbose: builtins.bool = ...,
        total: builtins.int = ...,
        failures: builtins.int = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["function",b"function","regex_pattern",b"regex_pattern","second_field",b"second_field","value",b"value","value_set",b"value_set"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["failures",b"failures","function",b"function","name",b"name","op",b"op","regex_pattern",b"regex_pattern","second_field",b"second_field","total",b"total","value",b"value","value_set",b"value_set","verbose",b"verbose"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["second_field",b"second_field"]) -> typing.Optional[typing_extensions.Literal["value","regex_pattern","value_set","function"]]: ...
global___ValueConstraintMsg = ValueConstraintMsg

class MultiColumnValueConstraintMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    NAME_FIELD_NUMBER: builtins.int
    DEPENDENT_COLUMNS_FIELD_NUMBER: builtins.int
    DEPENDENT_COLUMN_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    VALUE_SET_FIELD_NUMBER: builtins.int
    REFERENCE_COLUMNS_FIELD_NUMBER: builtins.int
    OP_FIELD_NUMBER: builtins.int
    VERBOSE_FIELD_NUMBER: builtins.int
    INTERNAL_DEPENDENT_COLUMNS_OP_FIELD_NUMBER: builtins.int
    TOTAL_FIELD_NUMBER: builtins.int
    FAILURES_FIELD_NUMBER: builtins.int
    name: typing.Text
    @property
    def dependent_columns(self) -> google.protobuf.struct_pb2.ListValue: ...
    dependent_column: typing.Text
    value: builtins.float
    @property
    def value_set(self) -> google.protobuf.struct_pb2.ListValue: ...
    @property
    def reference_columns(self) -> google.protobuf.struct_pb2.ListValue: ...
    op: global___Op.ValueType
    verbose: builtins.bool
    internal_dependent_columns_op: global___Op.ValueType
    total: builtins.int
    failures: builtins.int
    def __init__(self,
        *,
        name: typing.Text = ...,
        dependent_columns: typing.Optional[google.protobuf.struct_pb2.ListValue] = ...,
        dependent_column: typing.Text = ...,
        value: builtins.float = ...,
        value_set: typing.Optional[google.protobuf.struct_pb2.ListValue] = ...,
        reference_columns: typing.Optional[google.protobuf.struct_pb2.ListValue] = ...,
        op: global___Op.ValueType = ...,
        verbose: builtins.bool = ...,
        internal_dependent_columns_op: global___Op.ValueType = ...,
        total: builtins.int = ...,
        failures: builtins.int = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["dependent",b"dependent","dependent_column",b"dependent_column","dependent_columns",b"dependent_columns","reference",b"reference","reference_columns",b"reference_columns","value",b"value","value_set",b"value_set"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["dependent",b"dependent","dependent_column",b"dependent_column","dependent_columns",b"dependent_columns","failures",b"failures","internal_dependent_columns_op",b"internal_dependent_columns_op","name",b"name","op",b"op","reference",b"reference","reference_columns",b"reference_columns","total",b"total","value",b"value","value_set",b"value_set","verbose",b"verbose"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["dependent",b"dependent"]) -> typing.Optional[typing_extensions.Literal["dependent_columns","dependent_column"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["reference",b"reference"]) -> typing.Optional[typing_extensions.Literal["value","value_set","reference_columns"]]: ...
global___MultiColumnValueConstraintMsg = MultiColumnValueConstraintMsg

class ValueConstraintMsgs(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    CONSTRAINTS_FIELD_NUMBER: builtins.int
    MULTI_COLUMN_CONSTRAINTS_FIELD_NUMBER: builtins.int
    @property
    def constraints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ValueConstraintMsg]: ...
    @property
    def multi_column_constraints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MultiColumnValueConstraintMsg]: ...
    def __init__(self,
        *,
        constraints: typing.Optional[typing.Iterable[global___ValueConstraintMsg]] = ...,
        multi_column_constraints: typing.Optional[typing.Iterable[global___MultiColumnValueConstraintMsg]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["constraints",b"constraints","multi_column_constraints",b"multi_column_constraints"]) -> None: ...
global___ValueConstraintMsgs = ValueConstraintMsgs

class SummaryConstraintMsgs(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    CONSTRAINTS_FIELD_NUMBER: builtins.int
    @property
    def constraints(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___SummaryConstraintMsg]: ...
    def __init__(self,
        *,
        constraints: typing.Optional[typing.Iterable[global___SummaryConstraintMsg]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["constraints",b"constraints"]) -> None: ...
global___SummaryConstraintMsgs = SummaryConstraintMsgs

class DatasetConstraintMsg(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    class ValueConstraintsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___ValueConstraintMsgs: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___ValueConstraintMsgs] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    class SummaryConstraintsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor
        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: typing.Text
        @property
        def value(self) -> global___SummaryConstraintMsgs: ...
        def __init__(self,
            *,
            key: typing.Text = ...,
            value: typing.Optional[global___SummaryConstraintMsgs] = ...,
            ) -> None: ...
        def HasField(self, field_name: typing_extensions.Literal["value",b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing_extensions.Literal["key",b"key","value",b"value"]) -> None: ...

    PROPERTIES_FIELD_NUMBER: builtins.int
    VALUE_CONSTRAINTS_FIELD_NUMBER: builtins.int
    SUMMARY_CONSTRAINTS_FIELD_NUMBER: builtins.int
    TABLE_SHAPE_CONSTRAINTS_FIELD_NUMBER: builtins.int
    MULTI_COLUMN_VALUE_CONSTRAINTS_FIELD_NUMBER: builtins.int
    @property
    def properties(self) -> v0_messages_pb2.DatasetPropertiesV0: ...
    @property
    def value_constraints(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___ValueConstraintMsgs]: ...
    @property
    def summary_constraints(self) -> google.protobuf.internal.containers.MessageMap[typing.Text, global___SummaryConstraintMsgs]: ...
    @property
    def table_shape_constraints(self) -> global___SummaryConstraintMsgs: ...
    @property
    def multi_column_value_constraints(self) -> global___ValueConstraintMsgs: ...
    def __init__(self,
        *,
        properties: typing.Optional[v0_messages_pb2.DatasetPropertiesV0] = ...,
        value_constraints: typing.Optional[typing.Mapping[typing.Text, global___ValueConstraintMsgs]] = ...,
        summary_constraints: typing.Optional[typing.Mapping[typing.Text, global___SummaryConstraintMsgs]] = ...,
        table_shape_constraints: typing.Optional[global___SummaryConstraintMsgs] = ...,
        multi_column_value_constraints: typing.Optional[global___ValueConstraintMsgs] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["multi_column_value_constraints",b"multi_column_value_constraints","properties",b"properties","table_shape_constraints",b"table_shape_constraints"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["multi_column_value_constraints",b"multi_column_value_constraints","properties",b"properties","summary_constraints",b"summary_constraints","table_shape_constraints",b"table_shape_constraints","value_constraints",b"value_constraints"]) -> None: ...
global___DatasetConstraintMsg = DatasetConstraintMsg



import os
from abc import ABC, abstractmethod
from typing import Any, Optional, TypeVar

T = TypeVar("T", bound="Writer")


class Writable(ABC):
    @staticmethod
    def _safe_open_write(path):
        """Open `path` for writing, creating any parent directories as needed."""
        try:
            os.makedirs(os.path.dirname(path), exist_ok=True)
        except FileExistsError:
            pass
        return open(path, "w")

    @abstractmethod
    def get_default_path(self) -> str:
        pass

    @abstractmethod
    def write(self, path: Optional[str] = None, **kwargs: Any) -> None:
        pass


class Writer(ABC):
    # noinspection PyMethodMayBeStatic
    def check_interval(self, interval_seconds: int) -> None:
        """Validate an interval configuration for a given writer.

        Some writer only accepts certain interval configuration. By default, this should return True for a valid
        non-negative interval."""

    @abstractmethod
    def write(
        self,
        file: Writable,
        dest: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        pass

    @abstractmethod
    def option(self: T, **kwargs: Any) -> T:
        pass



from copy import deepcopy
from dataclasses import dataclass
from logging import getLogger
from typing import Callable, Dict, List, Optional, Tuple

from whylogs.core.metrics.metrics import Metric
from whylogs.core.view.column_profile_view import ColumnProfileView
from whylogs.core.view.dataset_profile_view import DatasetProfileView

logger = getLogger(__name__)


@dataclass
class MetricsSelector:
    metric_name: str
    column_name: Optional[str] = None  # <- TODO: return a collection of columns
    metrics_resolver: Optional[Callable[[DatasetProfileView], List[Metric]]] = None

    def column_profile(self, profile: DatasetProfileView) -> Optional[ColumnProfileView]:
        if profile is None:
            return None
        column_profile_view = profile.get_column(self.column_name)
        if column_profile_view is None:
            logger.warning(
                f"No column name {self.column_name} found in profile, available columns are: {profile.get_columns()}"
            )
        return column_profile_view

    def apply(self, profile: DatasetProfileView) -> List[Metric]:
        if self.metrics_resolver is not None:
            custom_result = self.metrics_resolver(profile)
            if isinstance(custom_result, List):
                return custom_result
            else:
                raise ValueError("metrics_resolver must return a list of Metrics if defined!")
        results: List[Metric] = []
        column_profile_view = self.column_profile(profile)
        if column_profile_view is None:
            return results
        metric = column_profile_view.get_metric(self.metric_name)
        if metric is None:
            logger.warning(
                f"{self.metric_name} not found in {self.column_name} available metrics are: "
                f"{column_profile_view.get_metric_component_paths()}"
            )
        else:
            results.append(metric)
        return results


@dataclass
class MetricConstraint:
    condition: Callable[[Metric], bool]
    name: str
    metric_selector: MetricsSelector
    require_column_existence: bool = True

    def _validate_metrics(self, metrics: List[Metric]) -> bool:
        for metric in metrics:
            if not self.condition(metric):
                return False
        return True

    def validate(self, dataset_profile: DatasetProfileView) -> bool:
        # custom metric resolver allows empty metrics
        if self.metric_selector is None:
            raise ValueError("can't call validate with an empty metric selector")
        metric_selector: MetricsSelector = self.metric_selector
        if metric_selector.metrics_resolver is not None:
            metrics = metric_selector.apply(dataset_profile)
            logger.info(f"validate using custom metric selector and found {metrics}")
            return self._validate_metrics(metrics)

        # standard selector requires a column be resolved
        column_profile = metric_selector.column_profile(dataset_profile)
        if column_profile is None:
            if self.require_column_existence:
                raise ValueError(
                    f"Could not get column {metric_selector.column_name}" f"from column {dataset_profile.get_columns()}"
                )
            else:
                logger.info(
                    f"validate could not find column {metric_selector.column_name} "
                    "but require_column_existence is false so returning True"
                )
                return True
        metrics = metric_selector.apply(dataset_profile)
        if metrics is None or len(metrics) == 0:
            if self.require_column_existence:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} so returning False. Available metrics on column are: "
                    f"{column_profile.get_metric_component_paths()}"
                )
                return False
            else:
                logger.info(
                    f"validate could not get metric {metric_selector.metric_name} from column "
                    f"{metric_selector.column_name} but require_column_existence is set so returning True. "
                    f"Available metrics on column are: {column_profile.get_metric_component_paths()}"
                )
                return True
        return self._validate_metrics(metrics)


class Constraints:
    column_constraints: Dict[str, Dict[str, MetricConstraint]]
    dataset_constraints: Dict[str, MetricConstraint]
    dataset_profile_view: Optional[DatasetProfileView]

    def __init__(
        self,
        dataset_profile_view: Optional[DatasetProfileView] = None,
        column_constraints: Optional[Dict[str, Dict[str, MetricConstraint]]] = None,
    ) -> None:
        if column_constraints is None:
            column_constraints = dict()
        self.column_constraints = column_constraints
        self.dataset_constraints = dict()
        self.dataset_profile_view = dataset_profile_view

    def validate(self, profile_view: Optional[DatasetProfileView] = None) -> bool:
        profile = self._resolve_profile_view(profile_view)
        column_names = self.column_constraints.keys()
        if len(column_names) == 0:
            logger.warning("validate was called with empty set of constraints, returning True!")
            return True

        for column_name in column_names:
            columnar_constraints = self.column_constraints[column_name]
            for constraint_name, metric_constraint in columnar_constraints.items():
                result = metric_constraint.validate(profile)
                if not result:
                    logger.info(f"{constraint_name} failed on column {column_name}")
                    return False
        metric_names = self.dataset_constraints.keys()
        for metric_name in metric_names:
            metric_constraint = self.dataset_constraints[metric_name]
            result = metric_constraint.validate(profile)
            if not result:
                logger.info(f"{metric_name} failed on dataset")
                return False
        return True

    def report(self, profile_view: Optional[DatasetProfileView] = None) -> List[Tuple[str, int, int]]:
        profile = self._resolve_profile_view(profile_view)
        column_names = self.column_constraints.keys()
        results: List[Tuple[str, int, int]] = []
        if len(column_names) == 0:
            logger.warning("report was called with empty set of constraints!")
            return results

        for column_name in column_names:
            columnar_constraints = self.column_constraints[column_name]
            for constraint_name, metric_constraint in columnar_constraints.items():
                result = metric_constraint.validate(profile)
                if not result:
                    logger.info(f"{constraint_name} failed on column {column_name}")
                    results.append((constraint_name, 0, 1))
                else:
                    results.append((constraint_name, 1, 0))
        metric_names = self.dataset_constraints.keys()
        for metric_name in metric_names:
            metric_constraint = self.dataset_constraints[metric_name]
            result = metric_constraint.validate(profile)
            if not result:
                logger.info(f"{metric_name} failed on dataset")
                results.append((metric_name, 0, 1))
            else:
                results.append((metric_name, 1, 0))
        return results

    def _resolve_profile_view(self, profile_view: Optional[DatasetProfileView]) -> DatasetProfileView:
        if profile_view is None and self.dataset_profile_view is None:
            raise ValueError(
                "Constraints need to be initialized on a profile or have a profile passed in before calling validate or report."
            )
        return profile_view if profile_view is not None else self.dataset_profile_view


class ConstraintsBuilder:
    def __init__(self, dataset_profile_view: DatasetProfileView, constraints: Optional[Constraints] = None) -> None:
        self._dataset_profile_view = dataset_profile_view
        if constraints is None:
            constraints = Constraints(dataset_profile_view=dataset_profile_view)
        self._constraints = constraints

    def get_metric_selectors(self) -> List[MetricsSelector]:
        selectors = []
        column_profiles = self._dataset_profile_view.get_columns()
        for column_name in column_profiles:
            column_profile = column_profiles[column_name]
            metric_component_paths = column_profile._metrics.keys()
            for metric_path in metric_component_paths:
                selectors.append(MetricsSelector(metric_name=metric_path, column_name=column_name))
        return selectors

    def add_constraint(self, constraint: MetricConstraint, ignore_missing: bool = False) -> "ConstraintsBuilder":
        # check that the column exists and we have a column profile view
        column_name = constraint.metric_selector.column_name
        column_profile_view = self._dataset_profile_view.get_column(column_name)
        if column_profile_view is None and not ignore_missing and constraint.metric_selector.metrics_resolver is None:
            raise ValueError(
                f"{column_name} was not found in set of this profile, the list of columns are: {self._dataset_profile_view.get_columns()}"
            )

        metric_selector = constraint.metric_selector
        metrics = metric_selector.apply(self._dataset_profile_view)
        if (metrics is None or len(metrics) == 0) and not (ignore_missing or column_name is None):
            raise ValueError(
                f"metrics not found for column {column_name}, available metric components are: {column_profile_view.get_metric_component_paths()}"
            )

        if column_name is None:
            self._constraints.dataset_constraints[constraint.name] = constraint
        else:
            if column_name not in self._constraints.column_constraints:
                self._constraints.column_constraints[column_name] = dict()
            self._constraints.column_constraints[column_name][constraint.name] = constraint

        return self

    def build(self) -> Constraints:
        return deepcopy(self._constraints)



from typing import Callable, Dict, Generic, Optional, TypeVar

import whylogs_sketching as ds  # type: ignore

from whylogs.core.metrics.decorators import DecoratedFunction, FuncType, _decorate_func

M = TypeVar("M")

_MAX_BUILT_IN_ID = 100


class _Aggregator(DecoratedFunction, Generic[M]):
    def __init__(self, *, func: FuncType, name: str):  # noqa
        self._func = func  # type: ignore
        self._name = name

    @property
    def name(self) -> Optional[str]:
        return self._name

    def __call__(self, lhs: M, rhs: M) -> M:
        return self._func(lhs, rhs)

    @classmethod
    def build(cls, func: FuncType, name: str) -> "_Aggregator":  # noqa
        return _Aggregator(func=func, name=name)


_TYPED_AGGREGATORS: Dict[type, _Aggregator] = {}
_ID_AGGREGATORS: Dict[int, _Aggregator] = {}


def _typed_aggregator(*, mtype: type, name: str) -> Callable[[Callable], _Aggregator]:
    decorated = _decorate_func(key=mtype, name=f"builtin.{name}", wrapper_dict=_TYPED_AGGREGATORS, clazz=_Aggregator)

    return decorated


def _id_aggregator(*, type_id: int, name: str):  # type: ignore
    if type_id <= 0:
        raise ValueError("Id must be a positive value")
    if type_id > _MAX_BUILT_IN_ID:
        raise ValueError("Built in id must be less than 100")

    decorated = _decorate_func(key=type_id, name=f"builtin.{name}", wrapper_dict=_ID_AGGREGATORS, clazz=_Aggregator)

    return decorated


NUM = TypeVar("NUM", float, int)


@_typed_aggregator(mtype=float, name="number")
@_typed_aggregator(mtype=int, name="number")
def _sum(lhs: NUM, rhs: NUM) -> NUM:
    return lhs + rhs


@_typed_aggregator(mtype=ds.kll_doubles_sketch, name="kll")
def _kll_merge(lhs: ds.kll_doubles_sketch, rhs: ds.kll_doubles_sketch) -> ds.kll_doubles_sketch:
    kll_copy = ds.kll_doubles_sketch(lhs)
    kll_copy.merge(rhs)

    return kll_copy


@_typed_aggregator(mtype=ds.hll_sketch, name="kll")
def _hll_merge(lhs: ds.hll_sketch, rhs: ds.hll_sketch) -> ds.hll_sketch:
    lg_k = max(lhs.lg_config_k, rhs.lg_config_k)
    tg_type = lhs.tgt_type
    if tg_type.value < rhs.tgt_type.value:
        tg_type = rhs.tgt_type

    copy = ds.hll_union(lg_k)
    copy.update(lhs)
    copy.update(rhs)

    return copy.get_result(tg_type)


@_typed_aggregator(mtype=ds.frequent_strings_sketch, name="kll")
def _fs_merge(lhs: ds.frequent_strings_sketch, rhs: ds.frequent_strings_sketch) -> ds.frequent_strings_sketch:
    copy = ds.frequent_strings_sketch.deserialize(lhs.serialize())
    copy.merge(rhs)

    return copy


class AggregatorRegistry:
    def __init__(self) -> None:
        self._id_aggs: Dict[int, _Aggregator] = _ID_AGGREGATORS.copy()
        self._typed_aggs = _TYPED_AGGREGATORS.copy()

    def get(self, *, type_id: int = 0, mtype: Optional[type] = None) -> Optional[_Aggregator]:
        if mtype is None and type_id <= 0:
            raise ValueError("id or metric component type must be set")
        result = None
        if type_id > 0:
            result = self._id_aggs.get(type_id)
        if result is None and mtype:
            result = self._typed_aggs.get(mtype)
        return result


_STANDARD_REGISTRY = AggregatorRegistry()


def aggregator(*, type_id: int, registry: Optional[AggregatorRegistry] = None):  # type: ignore
    if type_id < _MAX_BUILT_IN_ID:
        raise ValueError("Custom aggregator identifier must be equal or greater than 100")

    if registry is None:
        registry = _STANDARD_REGISTRY

    decorated = _decorate_func(key=type_id, name=f"custom.{type_id}", wrapper_dict=registry._id_aggs, clazz=_Aggregator)

    return decorated


def get_aggregator(
    *, type_id: int = 0, mtype: Optional[type] = None, registry: Optional[AggregatorRegistry] = None  # force kwargs
) -> Optional[_Aggregator]:
    if registry is None:
        registry = _STANDARD_REGISTRY

    return registry.get(type_id=type_id, mtype=mtype)



"""Centralized module for core math logic.

These utility functions can be hard to reason about so it's important to keep them as
centralized as possible and close to the original pseudo code.
"""
from collections import namedtuple

VarianceM2Result = namedtuple("VarianceM2Result", "n mean m2")


def parallel_variance_m2(*, first: VarianceM2Result, second: VarianceM2Result) -> VarianceM2Result:
    """
    Verbatim copy of the parallel variance algorithm from Wikipedia.

    Source: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
    Args:
        first: the :class:VarianceM2Result object for the first stream
        second: the :class:VarianceM2Result object for the second stream

    Returns:
        The variance result of the overall stream.

    """
    (n_a, avg_a, m2_a) = first
    (n_b, avg_b, m2_b) = second
    n = n_a + n_b
    delta = avg_b - avg_a
    mean = (avg_a * n_a + avg_b * n_b) / (n_a + n_b)
    m2 = m2_a + m2_b + delta**2 * n_a * n_b / n
    return VarianceM2Result(n=n, mean=mean, m2=m2)


def welford_online_variance_m2(*, existing: VarianceM2Result, new_value: float) -> VarianceM2Result:
    (n, mean, m2) = existing

    n += 1
    delta = new_value - mean
    mean += delta / n
    delta2 = new_value - mean
    m2 += delta * delta2
    return VarianceM2Result(n=n, mean=mean, m2=m2)



"""
A read/write library for length-delimited protobuf messages.

Based on: https://github.com/soulmachine/delimited-protobuf/blob/main/delimited_protobuf.py"""
from typing import IO, Type, TypeVar

from google.protobuf.message import Message

T = TypeVar("T", bound=Message)


def _read_varint(stream: IO[bytes], offset: int = 0) -> int:
    from google.protobuf.internal.decoder import _DecodeVarint  # type: ignore

    """Read a varint from the stream."""
    if offset > 0:
        stream.seek(offset)
    buf: bytes = stream.read(1)
    if buf == b"":
        return 0  # reached EOF
    while (buf[-1] & 0x80) >> 7 == 1:  # while the MSB is 1
        new_byte = stream.read(1)
        if new_byte == b"":
            raise EOFError("unexpected EOF")
        buf += new_byte
    varint, _ = _DecodeVarint(buf, 0)
    return varint


def read_delimited_protobuf(stream: IO[bytes], proto_class_name: Type[T], offset: int = 0) -> T:
    """Read a single length-delimited message from the given stream."""
    size = _read_varint(stream, offset=offset)
    if size == 0:
        return proto_class_name()
    buf = stream.read(size)
    msg = proto_class_name()
    msg.ParseFromString(buf)
    return msg


def write_delimited_protobuf(stream: IO[bytes], msg: T) -> None:
    """Write a single length-delimited message to the given stream."""
    assert stream is not None
    from google.protobuf.internal.encoder import _EncodeVarint  # type: ignore

    _EncodeVarint(stream.write, msg.ByteSize())
    stream.write(msg.SerializeToString())



import tempfile
from logging import getLogger
from typing import Any, Dict, Optional

from whylogs.api.writer.writer import Writable
from whylogs.core.proto import (
    ChunkHeader,
    ChunkMessage,
    ChunkOffsets,
    DatasetProfileHeader,
    DatasetProperties,
    DatasetSegmentHeader,
    MetricComponentMessage,
)
from whylogs.core.proto import Segment as SegmentMessage
from whylogs.core.segment import Segment
from whylogs.core.segmentation_partition import SegmentationPartition
from whylogs.core.utils import write_delimited_protobuf
from whylogs.core.utils.timestamp_calculations import to_utc_milliseconds
from whylogs.core.view.dataset_profile_view import (
    WHYLOGS_MAGIC_HEADER_BYTES,
    DatasetProfileView,
)
from whylogs.migration.converters import (
    _generate_segment_tags_metadata,
    v1_to_dataset_profile_message_v0,
)

logger = getLogger(__name__)

_BUFFER_CHUNK = 1024


class SegmentedDatasetProfileView(Writable):
    _profile_view: DatasetProfileView
    _segment: Segment
    _partition: SegmentationPartition

    def __init__(
        self,
        *,
        profile_view: DatasetProfileView,
        segment: Segment,
        partition: SegmentationPartition,
    ):
        self._profile_view = profile_view
        self._segment = segment
        self._partition = partition

    @property
    def segment(self) -> Segment:
        return self._segment

    @property
    def partition(self) -> SegmentationPartition:
        return self._partition

    @property
    def profile_view(self) -> DatasetProfileView:
        return self._profile_view

    @property
    def dataset_timestamp(self):
        return self.profile_view.dataset_timestamp

    @property
    def creation_timestamp(self):
        return self.profile_view.creation_timestamp

    def get_default_path(self) -> str:
        return f"profile_{self._profile_view.creation_timestamp}_{self._segment.parent_id}_{'_'.join(self._segment.key)}.bin"

    def _write_as_v0_message(self, path: Optional[str] = None, **kwargs: Any) -> None:
        message_v0 = v1_to_dataset_profile_message_v0(self.profile_view, self.segment, self.partition)
        path = path or self.get_default_path()
        with open(path, "w+b") as out_f:
            write_delimited_protobuf(out_f, message_v0)

    def _write_v1(self, path: Optional[str] = None, **kwargs: Any) -> None:
        all_metric_component_names = set()
        path = path or self.get_default_path()

        # capture the list of all metric component paths
        for col in self.profile_view._columns.values():
            all_metric_component_names.update(col.get_metric_component_paths())
        metric_name_list = list(all_metric_component_names)
        metric_name_list.sort()
        metric_name_indices: Dict[str, int] = {}
        metric_index_to_name: Dict[int, str] = {}
        for i in range(0, len(metric_name_list)):
            metric_name_indices[metric_name_list[i]] = i
            metric_index_to_name[i] = metric_name_list[i]

        column_chunk_offsets: Dict[str, ChunkOffsets] = {}
        with tempfile.TemporaryFile("w+b") as f:
            for col_name in sorted(self.profile_view._columns.keys()):
                column_chunk_offsets[col_name] = ChunkOffsets(offsets=[f.tell()])

                col = self.profile_view._columns[col_name]

                # for a given column, turn it into a ChunkMessage.
                indexed_component_messages: Dict[int, MetricComponentMessage] = {}
                metric_components = col.to_protobuf().metric_components
                for m_name, m_component in metric_components.items():
                    index = metric_name_indices.get(m_name)
                    if index is None:
                        raise ValueError(f"Missing metric from index mapping. Metric name: {m_name}")
                    indexed_component_messages[index] = m_component

                chunk_msg = ChunkMessage(metric_components=indexed_component_messages)
                chunk_header = ChunkHeader(type=ChunkHeader.ChunkType.COLUMN, length=chunk_msg.ByteSize())
                write_delimited_protobuf(f, chunk_header)
                f.write(chunk_msg.SerializeToString())

            total_len = f.tell()
            f.flush()

            # calculate segment tags based on columnar segments
            segment_message_tags, segment_tags, segment_metadata = _generate_segment_tags_metadata(
                self.segment, self.partition
            )

            if not segment_tags:
                raise NotImplementedError(
                    f"Serialization of segments requires segment tags but none calculated for partition: {self.partition} and segments: {self.segment}"
                )

            segment_message = SegmentMessage()
            segment_message.tags.extend(segment_tags)
            segments_message_field = [segment_message]

            properties = DatasetProperties(
                dataset_timestamp=to_utc_milliseconds(self.profile_view._dataset_timestamp),
                creation_timestamp=to_utc_milliseconds(self.profile_view._creation_timestamp),
                metadata=segment_metadata,
                tags=segment_message_tags,
            )

            logger.warn(f"constructed DatasetProperties for segmented profile file: {properties}")

            dataset_header = DatasetProfileHeader(
                column_offsets=column_chunk_offsets,
                properties=properties,
                length=total_len,
                indexed_metric_paths=metric_index_to_name,
            )

            # TODO: multi segment file format requires multiple offset calculations.
            #  Single segment file only supported initially.
            #  This creates the offsets dictionary with correct keys but 0 offset values, later update offsets
            #  based on file position offset in the temp file written, similarly to how Chunk
            segment_offsets: Dict[int, int] = {
                n: 0 for n in range(len(segments_message_field))
            }  # update this after writing

            # single file segments.
            dataset_segment_header = DatasetSegmentHeader(has_segments=True, offsets=segment_offsets)

            # TODO: calculate other segments offsets when we support multiple segments per file
            write_delimited_protobuf(f, dataset_segment_header)
            first_segment_offset = f.tell() - total_len
            dataset_segment_header.offsets[0] = first_segment_offset

            # only single segment files at first.
            dataset_segment_header.segments.extend(segments_message_field)

            with open(path, "w+b") as out_f:
                out_f.write(WHYLOGS_MAGIC_HEADER_BYTES)
                write_delimited_protobuf(out_f, dataset_segment_header)
                logger.debug(
                    f"Writing segmented profile file: whylogs file and segment headers wrote {out_f.tell()} bytes"
                )
                write_delimited_protobuf(out_f, dataset_header)
                logger.debug(
                    f"Writing segmented profile file: and with dataset header wrote a total of {out_f.tell()} bytes before writing the chunks."
                )

                f.seek(0)
                while f.tell() < total_len:
                    buffer = f.read(_BUFFER_CHUNK)
                    out_f.write(buffer)
                logger.debug(f"Writing segmented profile file: complete! total of {out_f.tell()} bytes written.")

    def write(self, path: Optional[str] = None, **kwargs: Any) -> None:
        if kwargs.get("use_v0"):
            self._write_as_v0_message(path, **kwargs)
        else:
            self._write_v1(path, **kwargs)



from .annotations import profiling
from .logger import (
    ResultSet,
    log,
    log_classification_metrics,
    log_regression_metrics,
    logger,
    read,
    reader,
    write,
)

__ALL__ = [
    log,
    log_classification_metrics,
    log_regression_metrics,
    logger,
    read,
    reader,
    write,
    profiling,
    ResultSet,
]



from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Dict, List


@dataclass
class BaseConfig:
    data_folder: str = "whylogs_data"
    description_folder: str = "whylogs.datasets.descr"


@dataclass
class DatasetConfig:
    folder_name: str
    url: str
    baseline_start_timestamp: Dict[str, datetime]
    inference_start_timestamp: Dict[str, datetime]
    max_interval: int
    base_unit: str
    available_versions: tuple
    target_columns: dict
    prediction_columns: dict
    miscellaneous_columns: dict
    ignore_columns: dict


@dataclass
class EcommerceConfig(DatasetConfig):
    folder_name: str = "ecommerce"
    description_file: str = "ecommerce.rst"
    available_versions: tuple = ("base",)
    # url: str = "./whylogs/datasets"
    url: str = "https://whylabs-public.s3.us-west-2.amazonaws.com/whylogs_examples/Ecommerce"
    baseline_start_timestamp: Dict[str, datetime] = field(
        default_factory=lambda: {
            "base": datetime(year=2022, month=8, day=9, tzinfo=timezone.utc),
        }
    )
    inference_start_timestamp: Dict[str, datetime] = field(
        default_factory=lambda: {
            "base": datetime(year=2022, month=8, day=19, tzinfo=timezone.utc),
        }
    )
    max_interval: int = 21
    base_unit: str = "D"
    target_columns: Dict[str, List[str]] = field(default_factory=lambda: {"base": ["output_discount"]})
    prediction_columns: Dict[str, List[str]] = field(
        default_factory=lambda: {
            "base": ["output_prediction", "output_score"],
        }
    )
    miscellaneous_columns: Dict[str, List[str]] = field(
        default_factory=lambda: {
            "base": [
                "category.Baby Care",
                "category.Bakery, Cakes and Dairy",
                "category.Beauty and Hygiene",
                "category.Beverages",
                "category.Cleaning and Household",
                "category.Eggs, Meat and Fish",
                "category.Foodgrains, Oil and Masala",
                "category.Fruits and Vegetables",
                "category.Gourmet and World Food",
                "category.Kitchen, Garden and Pets",
                "category.Snacks and Branded Foods",
            ],
        }
    )
    ignore_columns: Dict[str, List[str]] = field(
        default_factory=lambda: {
            "base": ["date"],
        }
    )


@dataclass
class WeatherConfig(DatasetConfig):
    folder_name: str = "weather_forecast"
    description_file: str = "weather.rst"
    url: str = "https://whylabs-public.s3.us-west-2.amazonaws.com/whylogs_examples/WeatherForecast"
    baseline_start_timestamp: Dict[str, datetime] = field(
        default_factory=lambda: {
            "in_domain": datetime(year=2018, month=9, day=1, tzinfo=timezone.utc),
            "out_domain": datetime(year=2018, month=9, day=1, tzinfo=timezone.utc),
        }
    )
    inference_start_timestamp: Dict[str, datetime] = field(
        default_factory=lambda: {
            "in_domain": datetime(year=2019, month=2, day=1, tzinfo=timezone.utc),
            "out_domain": datetime(year=2019, month=5, day=14, tzinfo=timezone.utc),
        }
    )

    max_interval: int = 30
    base_unit: str = "D"
    available_versions: tuple = ("in_domain", "out_domain")
    target_columns: Dict[str, List[str]] = field(
        default_factory=lambda: {"in_domain": ["temperature"], "out_domain": ["temperature"]}
    )
    prediction_columns: Dict[str, List[str]] = field(
        default_factory=lambda: {
            "in_domain": ["prediction_temperature", "uncertainty"],
            "out_domain": ["prediction_temperature", "uncertainty"],
        }
    )
    miscellaneous_columns: Dict[str, List[str]] = field(
        default_factory=lambda: {
            "in_domain": ["meta_latitude", "meta_longitude", "meta_climate"],
            "out_domain": ["meta_latitude", "meta_longitude", "meta_climate"],
        }
    )
    ignore_columns: Dict[str, List[str]] = field(
        default_factory=lambda: {
            "in_domain": ["date"],
            "out_domain": ["date"],
        }
    )



from .metric_constraints import (
    Constraints,
    ConstraintsBuilder,
    MetricConstraint,
    MetricsSelector,
)

__ALL__ = [
    Constraints,
    ConstraintsBuilder,
    MetricConstraint,
    MetricsSelector,
]



import atexit
import logging
import math
import os
import time
from datetime import datetime
from threading import Timer
from typing import Any, Callable, Dict, List, Optional

from typing_extensions import Literal

from whylogs.api.logger.logger import Logger
from whylogs.api.writer import Writer
from whylogs.core import DatasetProfile, DatasetProfileView, DatasetSchema
from whylogs.core.stubs import pd

logger = logging.getLogger(__name__)


class Scheduler(object):
    """
    Multithreading scheduler.

    Schedule a function to be called repeatedly based on a schedule.
    """

    _timer: Timer

    def __init__(self, initial: float, interval: float, function: Callable, *args: Any, **kwargs: Any):
        self.initial = initial
        self._ran_initial = False
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.start()

    def _run(self) -> None:
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self) -> None:
        if not self.is_running:
            interval = self.interval
            if not self._ran_initial:
                interval = self.initial
                self._ran_initial = True
            self._timer = Timer(interval, self._run)
            self._timer.start()
            self.is_running = True

    def stop(self) -> None:
        self._timer.cancel()
        self.is_running = False


class TimedRollingLogger(Logger):
    """A rolling logger that continuously rotates files based on time."""

    def __init__(
        self,
        schema: Optional[DatasetSchema] = None,
        *,
        base_name: Optional[str] = None,
        file_extension: Optional[str] = None,
        interval: int = 1,
        when: Literal["S", "M", "H", "D"] = "H",
        utc: bool = False,
        aligned: bool = True,
        fork: bool = False,
        skip_empty: bool = False,
        callback: Optional[Callable[[Writer, DatasetProfileView, str], None]] = None,
    ):
        super().__init__(schema)
        if base_name is None:
            base_name = "profile"
        if file_extension is None:
            file_extension = ".bin"

        self.file_extension = file_extension
        self.base_name = base_name
        self.aligned = aligned
        self.fork = fork
        self.skip_empty = skip_empty

        # base on TimedRotatingFileHandler
        self.when = when.upper()
        if self.when == "S":
            self.interval = 1  # one second
            self.suffix = "%Y-%m-%d_%H-%M-%S"
        elif self.when == "M":
            self.interval = 60  # one minute
            self.suffix = "%Y-%m-%d_%H-%M"
        elif self.when == "H":
            self.interval = 60 * 60  # one hour
            self.suffix = "%Y-%m-%d_%H"
        elif self.when == "D" or self.when == "MIDNIGHT":
            self.interval = 60 * 60 * 24  # one day
            self.suffix = "%Y-%m-%d"
        self.interval = self.interval * interval  # multiply by units requested
        self.utc = utc

        now = time.time()
        self._current_batch_timestamp = self._compute_current_batch_timestamp(now)

        self._current_profile: DatasetProfile = DatasetProfile(
            schema=schema,
            dataset_timestamp=datetime.utcfromtimestamp(self._current_batch_timestamp),
        )

        initial_run_after = (self._current_batch_timestamp + self.interval) - now
        if initial_run_after <= 0:
            logger.error(
                "Negative initial run after. This shouldn't happen so something went wrong with the clock here"
            )
            initial_run_after = self.interval
        self._callback = callback
        self._scheduler = Scheduler(initial_run_after, interval=self.interval, function=self._do_rollover)

        self._scheduler.start()

        atexit.register(self.close)

    def check_writer(self, writer: Writer) -> None:
        writer.check_interval(self.interval)

    def _compute_current_batch_timestamp(self, now: Optional[float] = None) -> int:
        if now is None:
            now = time.time()
        rounded_now = int(now)
        if self.aligned:
            return int(math.floor((rounded_now - 1) / self.interval)) * self.interval + self.interval
        return rounded_now

    def _get_matching_profiles(
        self, obj: Any = None, *, pandas: Optional[pd.DataFrame] = None, row: Optional[Dict[str, Any]] = None
    ) -> List[DatasetProfile]:
        return [self._current_profile]

    def _do_rollover(self) -> None:
        if self._is_closed:
            return
        old_profile = self._current_profile

        self._current_batch_timestamp = self._compute_current_batch_timestamp()
        dataset_timestamp = datetime.utcfromtimestamp(self._current_batch_timestamp)
        self._current_profile = DatasetProfile(
            schema=self._schema,
            dataset_timestamp=dataset_timestamp,
        )

        self._flush(old_profile)

    def _flush(self, profile: DatasetProfile) -> None:
        if profile is None:
            return
        if self.skip_empty and profile.is_empty:
            logger.debug("skip_empty is set. Skipping empty profiles")
            return

        pid = 0
        if self.fork:
            pid = os.fork()

        if pid > 0:
            logger.debug("Forked child process. Child process ID: %d", pid)
        else:
            if self.fork:
                logger.debug("In child process")
            else:
                logger.debug("Didn't fork. Writing in the same process")
            if self.utc:
                time_tuple = time.gmtime(self._current_batch_timestamp)
            else:
                time_tuple = time.localtime(self._current_batch_timestamp)
                current_time = int(time.time())

                dst_now = time.localtime(current_time)[-1]
                dst_then = time_tuple[-1]
                if dst_now != dst_then:
                    if dst_now:
                        addend = 3600
                    else:
                        addend = -3600
                    time_tuple = time.localtime(self._current_batch_timestamp + addend)
            timed_filename = f"{self.base_name}.{time.strftime(self.suffix, time_tuple)}{self.file_extension}"
            logging.debug("Writing out put with timed_filename: %s", timed_filename)

            while profile.is_active:
                time.sleep(1)

            for w in self._writers:
                w.write(profile=profile.view(), dest=timed_filename)
                if self._callback and callable(self._callback):
                    self._callback(w, profile.view(), timed_filename)

    def close(self) -> None:
        logging.debug("Closing the writer")
        if not self._is_closed:
            self._scheduler.stop()
            self._do_rollover()
        self._is_closed: bool = True



import html
import json
import logging
import os
from typing import Any, Dict, Optional

from IPython.core.display import HTML  # type: ignore

from whylogs.api.usage_stats import emit_usage
from whylogs.core.configs import SummaryConfig
from whylogs.core.constraints import Constraints
from whylogs.core.view.dataset_profile_view import DatasetProfileView
from whylogs.viz.enums.enums import PageSpec, PageSpecEnum
from whylogs.viz.utils.frequent_items_calculations import zero_padding_frequent_items
from whylogs.viz.utils.html_template_utils import _get_compiled_template
from whylogs.viz.utils.profile_viz_calculations import (
    add_feature_statistics,
    frequent_items_from_view,
    generate_profile_summary,
    generate_summaries,
    histogram_from_view,
)

logger = logging.getLogger(__name__)
emit_usage("visualizer")


class NotebookProfileVisualizer:
    """
    Visualize and compare profiles for drift detection, data quality, distribution comparison and feature statistics.

    NotebookProfileVisualizer enables visualization features for Jupyter Notebook environments, but also enables
    download
    of the generated reports as HTML files.

    Examples
    --------

    Create target and reference dataframes:

    .. code-block:: python

        import pandas as pd

        data_target = {
            "animal": ["cat", "hawk", "snake", "cat", "snake", "cat", "cat", "snake", "hawk","cat"],
            "legs": [4, 2, 0, 4, 0, 4, 4, 0, 2, 4],
            "weight": [4.3, None, 2.3, 7.8, 3.7, 2.5, 5.5, 3.3, 0.6, 13.3],
        }

        data_reference = {
            "animal": ["hawk", "hawk", "snake", "hawk", "snake", "snake", "cat", "snake", "hawk","snake"],
            "legs": [2, 2, 0, 2, 0, 0, 4, 0, 2, 0],
            "weight": [2.7, None, 1.2, 10.5, 2.2, 4.6, 3.8, 4.7, 0.6, 11.2],
        }

        target_df = pd.DataFrame(data_target)
        reference_df = pd.DataFrame(data_reference)

    Log data and create profile views:

    .. code-block:: python

        import whylogs as why

        results = why.log(pandas=target_df)
        prof_view = results.view()

        results_ref = why.log(pandas=reference_df)
        prof_view_ref = results_ref.view()

    Log data and create profile views:

    .. code-block:: python

        import whylogs as why

        results = why.log(pandas=target_df)
        prof_view = results.view()

        results_ref = why.log(pandas=reference_df)
        prof_view_ref = results_ref.view()

    Instantiate and set profile views:

    .. code-block:: python

        from whylogs.viz import NotebookProfileVisualizer

        visualization = NotebookProfileVisualizer()
        visualization.set_profiles(target_profile_view=prof_view,reference_profile_view=prof_view_ref)
    """

    _ref_view: Optional[DatasetProfileView]
    _target_view: DatasetProfileView

    @staticmethod
    def _display(template: str, page_spec: PageSpec, height: Optional[str]) -> "HTML":
        if not height:
            height = page_spec.height
        iframe = f"""<div></div><iframe srcdoc="{html.escape(template)}" width=100% height={height}
        frameBorder=0></iframe>"""
        display = HTML(iframe)
        return display

    def _display_distribution_chart(
        self, feature_name: str, difference: bool, cell_height: str = None, config: Optional[SummaryConfig] = None
    ) -> Optional[HTML]:
        if config is None:
            config = SummaryConfig()
        if difference:
            page_spec = PageSpecEnum.DIFFERENCED_CHART.value
        else:
            page_spec = PageSpecEnum.DISTRIBUTION_CHART.value

        template = _get_compiled_template(page_spec.html)
        if self._target_view:
            target_profile_features: Dict[str, Dict[str, Any]] = {feature_name: {}}
            reference_profile_features: Dict[str, Dict[str, Any]] = {feature_name: {}}

            target_column_profile_view = self._target_view.get_column(feature_name)
            if not target_column_profile_view:
                raise ValueError("ColumnProfileView for feature {} not found.".format(feature_name))

            target_profile_features[feature_name]["frequentItems"] = frequent_items_from_view(
                target_column_profile_view, feature_name, config
            )
            if self._ref_view:
                ref_col_view = self._ref_view.get_column(feature_name)
                if not ref_col_view:
                    raise ValueError("ColumnProfileView for feature {} not found.".format(feature_name))

                reference_profile_features[feature_name]["frequentItems"] = frequent_items_from_view(
                    ref_col_view, feature_name, config
                )

                (
                    target_profile_features[feature_name]["frequentItems"],
                    reference_profile_features[feature_name]["frequentItems"],
                ) = zero_padding_frequent_items(
                    target_feature_items=target_profile_features[feature_name]["frequentItems"],
                    reference_feature_items=reference_profile_features[feature_name]["frequentItems"],
                )
            else:
                logger.warning("Reference profile not detected. Plotting only for target feature.")
                reference_profile_features[feature_name]["frequentItems"] = [
                    {"value": x["value"], "estimate": 0} for x in target_profile_features[feature_name]["frequentItems"]
                ]  # Getting the same frequent items categories for target and adding 0 as estimate.
            distribution_chart = template(
                {
                    "profile_from_whylogs": json.dumps(target_profile_features),
                    "reference_profile_from_whylogs": json.dumps(reference_profile_features),
                }
            )
            result = self._display(distribution_chart, page_spec, cell_height)
            return result

        else:
            logger.warning("This method has to get at least a target profile, with valid feature title")
            return None

    def _display_histogram_chart(self, feature_name: str, cell_height: str = None) -> Optional[HTML]:
        page_spec = PageSpecEnum.DOUBLE_HISTOGRAM.value
        template = _get_compiled_template(page_spec.html)
        if self._target_view:
            target_features: Dict[str, Dict[str, Any]] = {feature_name: {}}
            ref_features: Dict[str, Dict[str, Any]] = {feature_name: {}}

            target_col_view = self._target_view.get_column(feature_name)
            if not target_col_view:
                raise ValueError(f"ColumnProfileView for feature {feature_name} not found.")

            target_histogram = histogram_from_view(target_col_view, feature_name)
            if self._ref_view:
                reference_column_profile_view = self._ref_view.get_column(feature_name)
                if not reference_column_profile_view:
                    raise ValueError("ColumnProfileView for feature {} not found.".format(feature_name))
                ref_histogram = histogram_from_view(reference_column_profile_view, feature_name)
            else:
                logger.warning("Reference profile not detected. Plotting only for target feature.")
                ref_histogram = target_histogram.copy()
                ref_histogram["counts"] = [
                    0 for _ in ref_histogram["counts"]
                ]  # To plot single profile, zero counts for non-existing profile.

            ref_features[feature_name]["histogram"] = ref_histogram
            target_features[feature_name]["histogram"] = target_histogram
            histogram_chart = template(
                {
                    "profile_from_whylogs": json.dumps(target_features),
                    "reference_profile_from_whylogs": json.dumps(ref_features),
                }
            )
            return self._display(histogram_chart, page_spec, height=cell_height)
        else:
            logger.warning("This method has to get at least a target profile, with valid feature title")
            return None

    def set_profiles(
        self, target_profile_view: DatasetProfileView, reference_profile_view: Optional[DatasetProfileView] = None
    ) -> None:
        """Set profiles for Visualization/Comparison.

        Drift calculation is done if both `target_profile` and `reference profile` are passed.

        Parameters
        ----------
        target_profile_view: DatasetProfileView, required
            Target profile to visualize.
        reference_profile_view: DatasetProfileView, optional
            Reference, or baseline, profile to be compared against the target profile.

        """
        self._target_view = target_profile_view
        self._ref_view = reference_profile_view

    def profile_summary(self, cell_height: str = None) -> HTML:
        page_spec = PageSpecEnum.PROFILE_SUMMARY.value
        template = _get_compiled_template(page_spec.html)

        try:
            profile_summary = generate_profile_summary(self._target_view, config=None)
            rendered_template = template(profile_summary)
            return self._display(rendered_template, page_spec, cell_height)
        except ValueError as e:
            logger.error("This method has to get target Dataset Profile View")
            raise e

    def summary_drift_report(self, height: Optional[str] = None) -> HTML:
        """Generate drift report between target and reference profiles.

        KS test is applied for continuous variables and ChiSquared test for categorical variables.
        If feature is missing from any profile, it will not be included in the report.
        Both target_profile_view and reference_profile_view must be set previously with `set_profiles`.

        Parameters
        ----------
        height: str, optional
            Preferred height, in pixels, for in-notebook visualization. Example:
            `"1000px"`. (Default is None)

        Returns
        -------
        HTML
            HTML Page of the given plot.

        Examples
        --------

        Generate Summary Drift Report (after setting profiles with `set_profiles`):

        .. code-block:: python

            visualization.summary_drift_report()

        """
        if not self._target_view or not self._ref_view:
            logger.error("This method has to get both target and reference profiles")
            raise ValueError

        page_spec = PageSpecEnum.SUMMARY_REPORT.value
        template = _get_compiled_template(page_spec.html)

        profiles_summary = generate_summaries(self._target_view, self._ref_view, config=None)
        rendered_template = template(profiles_summary)
        summary_drift_report = self._display(rendered_template, page_spec, height)
        return summary_drift_report

    def double_histogram(self, feature_name: str, cell_height: Optional[str] = None) -> HTML:
        """Plot overlayed histograms for specified feature present in both `target_profile` and `reference_profile`.

        Applicable to numerical features only.
        If reference profile was not set, `double_histogram` will plot single histogram for target profile.

        Parameters
        ----------
        feature_name: str
            Name of the feature to generate histograms.
        cell_height: str, optional
            Preferred cell height, in pixels, for in-notebook visualization. Example:
            `"1000px"`. (Default is None)

        Examples
        --------

        Generate double histogram plot for feature named `weight` (after setting profiles with `set_profiles`)

        .. code-block:: python

            visualization.double_histogram(feature_name="weight")
        """
        double_histogram = self._display_histogram_chart(feature_name, cell_height)
        return double_histogram

    def distribution_chart(self, feature_name: str, cell_height: Optional[str] = None) -> HTML:
        """Plot overlayed distribution charts for specified feature between two profiles.

        Applicable to categorical features.
        If reference profile was not set, `distribution_chart` will plot single chart for target profile.


        Parameters
        ----------
        feature_name : str
            Name of the feature to plot chart.
        cell_height : str, optional
            Preferred cell height, in pixels, for in-notebook visualization. Example:
            `cell_height="1000px"`. (Default is None)

        Returns
        -------
        HTML
            HTML Page of the given plot.

        Examples
        --------

        Generate distribution chart for `animal` feature (after setting profiles with `set_profiles`):

        .. code-block:: python

            visualization.distribution_chart(feature_name="animal")
        """
        difference = False
        distribution_chart = self._display_distribution_chart(feature_name, difference, cell_height)
        return distribution_chart

    def difference_distribution_chart(self, feature_name: str, cell_height: Optional[str] = None) -> HTML:
        """Plot overlayed distribution charts of differences between the categories of both profiles.

        Applicable to categorical features.

        Parameters
        ----------
        feature_name : str
            Name of the feature to plot chart.
        cell_height : str, optional
            Preferred cell height, in pixels, for in-notebook visualization. Example:
            `cell_height="1000px"`. (Default is None)

        Returns
        -------
        HTML
            HTML Page of the given plot.

        Examples
        --------

        Generate Difference Distribution Chart for feature named "animal":

        .. code-block:: python

            visualization.difference_distribution_chart(feature_name="animal")

        """
        difference = True
        difference_distribution_chart = self._display_distribution_chart(feature_name, difference, cell_height)
        return difference_distribution_chart

    def constraints_report(self, constraints: Constraints, cell_height: Optional[str] = None) -> HTML:
        page_spec = PageSpecEnum.CONSTRAINTS_REPORT.value
        template = _get_compiled_template(page_spec.html)
        rendered_template = template({"constraints_report": json.dumps(constraints.report())})
        constraints_report = self._display(rendered_template, page_spec, cell_height)
        return constraints_report

    def feature_statistics(
        self, feature_name: str, profile: str = "reference", cell_height: Optional[str] = None
    ) -> HTML:
        """
        Generate a report for the main statistics of specified feature, for a given profile (target or reference).

        Statistics include overall metrics such as distinct and missing values, as well as quantile and descriptive
        statistics.
        If `profile` is not passed, the default is the reference profile.

        Parameters
        ----------
        feature_name: str
            Name of the feature to generate histograms.
        profile: str
            Profile to be used to generate the report. (Default is `reference`)
        cell_height: str, optional
            Preferred cell height, in pixels, for in-notebook visualization. Example:
            `cell_height="1000px"`. (Default is None)

        Examples
        --------

        Generate Difference Distribution Chart for feature named "weight", for target profile:

        .. code-block:: python

            visualization.feature_statistics(feature_name="weight", profile="target")

        """
        page_spec = PageSpecEnum.FEATURE_STATISTICS.value
        template = _get_compiled_template(page_spec.html)
        if self._ref_view and profile.lower() == "reference":
            selected_profile_column = self._ref_view.get_column(feature_name)
        else:
            selected_profile_column = self._target_view.get_column(feature_name)

        rendered_template = template(
            {
                "profile_feature_statistics_from_whylogs": json.dumps(
                    add_feature_statistics(feature_name, selected_profile_column)
                )
            }
        )
        feature_statistics = self._display(rendered_template, page_spec, cell_height)
        return feature_statistics

    @staticmethod
    def write(
        rendered_html: HTML,
        preferred_path: Optional[str] = None,  # type: ignore
        html_file_name: Optional[str] = None,  # type: ignore
    ) -> None:
        """Create HTML file for a given report.

        Parameters
        ----------
        rendered_html: HTML, optional
            Rendered HTML returned by a given report.
        preferred_path: str, optional
            Preferred path to write the HTML file.
        html_file_name: str, optional
            Name for the created HTML file. If none is passed, created HTML will be named `ProfileVisualizer.html`

        Examples
        --------
        Dowloads an HTML page named `test.html` into the current working directory, with feature statistics for `weight` feature for the target profile.

        .. code-block:: python

            import os
            visualization.write(
                rendered_html=visualization.feature_statistics(feature_name="weight", profile="target"),
                html_file_name=os.getcwd() + "/test",
            )


        """
        if not html_file_name:
            html_file_name = "ProfileVisualizer"
        if preferred_path:
            full_path = os.path.join(os.path.expanduser(preferred_path), str(html_file_name) + ".html")
        else:
            full_path = os.path.join(os.pardir, "html_reports", str(html_file_name) + ".html")

        with open(os.path.abspath(full_path), "w") as saved_html:
            saved_html.write(rendered_html.data)



from typing import Dict, Optional

import numpy as np
from scipy import stats  # type: ignore
from typing_extensions import TypedDict
from whylogs_sketching import kll_doubles_sketch  # type: ignore

from whylogs.core.view.dataset_profile_view import DatasetProfileView  # type: ignore
from whylogs.viz.utils.frequent_items_calculations import (
    FrequentStats,
    get_frequent_stats,
    zero_padding_frequent_items,
)

QUANTILES = [0.0, 0.01, 0.05, 0.25, 0.5, 0.75, 0.95, 0.99, 1.0]


class ColumnDriftValue(TypedDict):
    """p-value for applied statistical test, along with the name of the applied test."""

    p_value: float
    test: str


def _compute_ks_test_p_value(
    target_distribution: kll_doubles_sketch, reference_distribution: kll_doubles_sketch
) -> Optional[ColumnDriftValue]:
    """Compute the Kolmogorov-Smirnov test p-value of two continuous distributions.

    Uses the quantile values and the corresponding CDFs to calculate the approximate KS statistic.
    Only applicable to continuous distributions.
    The null hypothesis expects the samples to come from the same distribution.

    Parameters
    ----------
    target_distribution : datasketches.kll_floats_sketch
        A kll_floats_sketch (quantiles sketch) from the target distribution's values
    reference_distribution : datasketches.kll_floats_sketch
        A kll_floats_sketch (quantiles sketch) from the reference (expected) distribution's values
        Can be generated from a theoretical distribution, or another sample for the same feature.

    Returns
    -------
        p_value : float
        The estimated p-value from the parametrized KS test, applied on the target and reference distributions'
        kll_floats_sketch summaries

    """
    if reference_distribution.is_empty() or target_distribution.is_empty():
        return None

    D_max = 0
    target_quantile_values = target_distribution.get_quantiles(QUANTILES)
    ref_quantile_values = reference_distribution.get_quantiles(QUANTILES)

    num_quantiles = len(QUANTILES)
    i, j = 0, 0
    while i < num_quantiles and j < num_quantiles:
        if target_quantile_values[i] < ref_quantile_values[j]:
            current_quantile = target_quantile_values[i]
            i += 1
        else:
            current_quantile = ref_quantile_values[j]
            j += 1

        cdf_target = target_distribution.get_cdf([current_quantile])[0]
        cdf_ref = reference_distribution.get_cdf([current_quantile])[0]
        D = abs(cdf_target - cdf_ref)
        if D > D_max:
            D_max = D

    while i < num_quantiles:
        cdf_target = target_distribution.get_cdf([target_quantile_values[i]])[0]
        cdf_ref = reference_distribution.get_cdf([target_quantile_values[i]])[0]
        D = abs(cdf_target - cdf_ref)
        if D > D_max:
            D_max = D
        i += 1

    while j < num_quantiles:
        cdf_target = target_distribution.get_cdf([ref_quantile_values[j]])[0]
        cdf_ref = reference_distribution.get_cdf([ref_quantile_values[j]])[0]
        D = abs(cdf_target - cdf_ref)
        if D > D_max:
            D_max = D
        j += 1

    m, n = sorted([target_distribution.get_n(), reference_distribution.get_n()], reverse=True)
    en = m * n / (m + n)

    p_value = stats.distributions.kstwo.sf(D_max, np.round(en))

    return {"p_value": p_value, "test": "ks"}


def _get_ks_p_value(target_view_column, reference_view_column) -> Optional[ColumnDriftValue]:
    target_dist_metric = target_view_column.get_metric("distribution")
    ref_dist_metric = reference_view_column.get_metric("distribution")

    if target_dist_metric is None or ref_dist_metric is None:
        return None

    target_kll_sketch = target_dist_metric.kll.value
    ref_kll_sketch = ref_dist_metric.kll.value

    ks_p_value = _compute_ks_test_p_value(target_kll_sketch, ref_kll_sketch)
    return ks_p_value


def _compute_chi_squared_test_p_value(
    target_distribution: FrequentStats, reference_distribution: FrequentStats
) -> Optional[ColumnDriftValue]:
    """
    Calculate the Chi-Squared test p-value for two discrete distributions.

    Uses the top frequent items summary, unique count estimate and total count estimate for each feature,
    to calculate the estimated Chi-Squared statistic.
    Applicable only to discrete distributions.

    Parameters
    ----------
    target_distribution : ReferenceDistributionDiscreteMessage
        The summary message of the target feature's distribution.
        Should be a ReferenceDistributionDiscreteMessage containing the frequent items,
        unique, and total count summaries.
    reference_distribution : ReferenceDistributionDiscreteMessage
        The summary message of the reference feature's distribution.
        Should be a ReferenceDistributionDiscreteMessage containing the frequent items,
        unique, and total count summaries.

    Returns
    -------
        p_value : ColumnDriftValue or None. ColumnDriftValue has fields `p-value` with the test's result,
        and the name of the test applied in the `test` field (chi-squared).
        The estimated p-value from the Chi-Squared test, applied on the target and reference distributions'
        frequent and unique items summaries
    """
    target_freq_items = target_distribution["frequent_items"]
    ref_freq_items = reference_distribution["frequent_items"]
    target_total_count = target_distribution["total_count"]
    target_unique_count = target_distribution["unique_count"]
    ref_total_count = reference_distribution["total_count"]

    if ref_total_count <= 0 or target_total_count <= 0:
        return None

    target_freq_items, ref_freq_items = zero_padding_frequent_items(target_freq_items, ref_freq_items)

    ref_dist_items = dict()
    for item in reference_distribution["frequent_items"]:
        ref_dist_items[item["value"]] = item["estimate"]
    proportion_ref_dist_items = {k: v / ref_total_count for k, v in ref_dist_items.items()}

    chi_sq = 0.0
    for item in target_freq_items:
        target_frequency = item["estimate"]
        if item["value"] in ref_dist_items:
            ref_frequency = int(proportion_ref_dist_items[item["value"]] * target_total_count)
        else:
            ref_frequency = 0

        if ref_frequency == 0:
            chi_sq = np.inf
            break
        chi_sq += (target_frequency - ref_frequency) ** 2 / ref_frequency

    degrees_of_freedom = target_unique_count - 1
    degrees_of_freedom = degrees_of_freedom if degrees_of_freedom > 0 else 1
    p_value = stats.chi2.sf(chi_sq, degrees_of_freedom)
    return {"p_value": p_value, "test": "chi-squared"}


def _get_chi2_p_value(target_view_column, reference_view_column) -> Optional[ColumnDriftValue]:
    target_frequent_stats: FrequentStats = get_frequent_stats(target_view_column, config=None)
    ref_frequent_stats: FrequentStats = get_frequent_stats(reference_view_column, config=None)

    if not target_frequent_stats or not ref_frequent_stats:
        return None

    chi2_p_value = _compute_chi_squared_test_p_value(
        target_distribution=target_frequent_stats, reference_distribution=ref_frequent_stats
    )
    return chi2_p_value


def calculate_drift_values(
    target_view: DatasetProfileView, reference_view: DatasetProfileView
) -> Dict[str, Optional[ColumnDriftValue]]:
    """Calculate drift values between both profiles. Applicable for numerical and categorical features.

    Calculates drift only for features found in both profiles, and ignore those not found in either profile.

    Parameters
    ----------
    target_view : DatasetProfileView
        Target Profile View
    reference_view : DatasetProfileView
        Reference Profile View

    Returns
    -------
    drift_values: Dict[str, Optional[ColumnDriftValue]]
        A dictionary of the p-values, along with the type of test applied, for the given features.
    """
    drift_values: Dict[str, Optional[ColumnDriftValue]] = {}
    target_view_columns = target_view.get_columns()
    reference_view_columns = reference_view.get_columns()
    for target_column_name in target_view_columns:
        if target_column_name in reference_view_columns:
            target_view_column = target_view_columns[target_column_name]
            reference_view_column = reference_view_columns[target_column_name]

            drift_values[target_column_name] = _get_ks_p_value(
                target_view_column=target_view_column, reference_view_column=reference_view_column
            ) or _get_chi2_p_value(target_view_column=target_view_column, reference_view_column=reference_view_column)

    return drift_values



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import date, datetime
from typing import Iterable, List, Optional, Union

import pandas as pd

from whylogs.datasets.configs import DatasetConfig

logger = logging.getLogger(__name__)


@dataclass(init=False)
class Batch:
    """Batch object that encapsulate data and information for a given batch.

    - timestamp: the batch's timestamp (at the start)
    - data: the complete dataframe
    - features: input features
    - target: output feature(s)
    - prediction: output prediction and, possibly, features such as uncertainty, confidence, probability
    - misc: metadata features that are not of any of the previous categories, but still contain relevant information about the data.

    A batch can represent either a baseline or inference batch.
    The complete data is a sum of the remaining dataframe properties: features, target, prediction, and misc.
    """

    def __init__(self, timestamp: date, data: pd.DataFrame, dataset_config: DatasetConfig, version: str):
        self._timestamp = timestamp
        self._data = data
        self.dataset_config = dataset_config
        self.version = version

    @property
    def data(self) -> pd.DataFrame:
        """The complete dataframe for all available features."""

        data_cols = list(self._data.columns)
        ignore_cols = list(self.dataset_config.ignore_columns[self.version])
        return self._data[[col for col in data_cols if col not in ignore_cols]]

    @property
    def timestamp(self) -> date:
        """The batch's timestamp (at the start)"""
        return self._timestamp

    @property
    def target(self) -> pd.DataFrame:
        """Ouput feature(s)"""
        return self._data[list(self.dataset_config.target_columns[self.version])]

    @property
    def prediction(self) -> pd.DataFrame:
        """Output prediction and, possibly, features such as uncertainty, confidence, probability scores"""
        return self._data[list(self.dataset_config.prediction_columns[self.version])]

    @property
    def extra(self) -> pd.DataFrame:
        """Metadata features that are not of any of the previous categories, but still contain relevant information about the data."""
        return self._data[list(self.dataset_config.miscellaneous_columns[self.version])]

    @property
    def features(self) -> pd.DataFrame:
        """Input features"""
        data_cols = list(self._data.columns)
        target_cols = list(self.dataset_config.target_columns[self.version])
        prediction_cols = list(self.dataset_config.prediction_columns[self.version])
        misc_cols = list(self.dataset_config.miscellaneous_columns[self.version])
        ignore_cols = list(self.dataset_config.ignore_columns[self.version])
        return self._data[
            [col for col in data_cols if col not in target_cols + prediction_cols + misc_cols + ignore_cols]
        ]


class Dataset(ABC):
    """Abstract class representing a dataset."""

    @classmethod
    @abstractmethod
    def describe_versions(cls) -> List[str]:
        """Describe available versions for the given dataset."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def describe(cls) -> str:
        """Display overall dataset description."""
        raise NotImplementedError

    @abstractmethod
    def set_parameters(
        self,
        inference_interval: str,
        baseline_timestamp: Optional[Union[date, datetime]] = None,
        inference_start_timestamp: Optional[Union[date, datetime]] = None,
        original: Optional[bool] = None,
    ) -> None:
        """Set interval and timestamp parameters for the dataset object.

        Parameters
        ----------
        inference_interval : str
            Time period for each batch retrieved from the inference dataset. E.g. daily batches would be set as "1d"
        baseline_timestamp : Optional[Union[date, datetime]], optional
            The timestamp for the baseline dataset. Will be set to the dataset's original timestamp if original=True. By default None
        inference_start_timestamp : Optional[Union[date, datetime]], optional
            The timestamp for the start of the inference dataset. Will be set to the dataset's original timestamp if original=True. By default None
        original: Optional[bool], optional
            If true, will set both baseline_timestamp and inference_start_timestamp to its original values.
        """
        raise NotImplementedError

    @abstractmethod
    def get_baseline(self) -> Batch:
        """Get baseline Batch object.

        Returns
        -------
        Batch
            A batch object representing the complete baseline data.
        """
        raise NotImplementedError

    @abstractmethod
    def get_inference_data(
        self, target_date: Optional[Union[date, datetime]] = None, number_batches: Optional[int] = None
    ) -> Union[Batch, Iterable[Batch]]:
        """Get inference batch(es)

        Parameters
        ----------
        target_date : Optional[Union[date, datetime]], optional
            If target_date is set, a single batch will be returned for the given date(or datetime).
            If both target_date and number_batches are defined, an error will be raised.
        number_batches : Optional[int], optional
            If number_batches is set to `n`, an iterator of `n` inference batches will be returned, starting from `inference_start_timestamp`.
            If both target_date and number_batches are defined, an error will be raised.

        Returns
        -------
        Union[Batch, Iterable[Batch]]
            Can return either a single or multiple batches, according to the parameters passed.
        """
        raise NotImplementedError



from .protobuf_utils import read_delimited_protobuf, write_delimited_protobuf
from .stats_calculations import get_distribution_metrics
from .utils import deprecated_alias

__ALL__ = [read_delimited_protobuf, write_delimited_protobuf, get_distribution_metrics, deprecated_alias]
